from __future__ import print_function
#import matplotlib
#matplotlib.use('Agg') # prevents crashes if no X server present
from builtins import str
from builtins import range
from yambopy import *
import sys
import os
import argparse
import re
import numpy as np
import matplotlib.pyplot as plt

"""
After real-time calculations are completed, plots the Transient Absorption (TA) at different times.
Setting inside the script : prefix
"""

parser = argparse.ArgumentParser(description='Make Transiant Absorption (TA) plots from RT-BSE')
parser.add_argument('-f' ,'--folder'    , help='Real-time folder (e.g.: rt-24x24)')
parser.add_argument('-j' ,'--job'       , help='Name of job (e.g.: QSSIN-2.0eV)')
parser.add_argument('-p' ,'--prefix'    , help='Prefix of the BSE calculations (e.g.: B-XRK-XG)')
parser.add_argument('-nt','--notext'    , help='Skips the writing of the data', action='store_false')
args = parser.parse_args()

folder = args.folder
job = args.job
prefix = args.prefix

source = folder+'/'+job

print('Packing relevant calculations')
pack_files_in_folder(source,mask=prefix)
print('Done.')

data = YamboAnalyser(source)
output = data.get_data((prefix,'eps'))

# keys to read the outputs in order
keys=list(output.keys())

# sorting
s=[]
for i,key in enumerate(keys):
        s.append((int(re.search("\d{1,}", key).group()),key))

s=sorted(s) # s is list of tuples (int,str)
            # having integers allows to sort correctly

times = ['t'+str(i[0]) for i in s]
keys  = [i[1] for i in s]

print("Sorted keys: ",keys)

# output[key][n][0] is x
# output[key][n][1] is y

# Number of lines (number of x points)
nlines = len(output[keys[0]])
# We want to create a file that is x, y1, y2, ..., yn
array = np.zeros((nlines,len(keys)+1))
diff = np.zeros((nlines,len(keys)))

for l in range(0,nlines):
    # first, value of x
    x = output[keys[0]][l][0]
    array[l][0]=x
    diff[l][0]=x
    # then all the y's
    # t0
    y0 = output[keys[0]][l][1]
    array[l][1] = y0
    # additional columns
    for i,key in enumerate(keys):
        # t_i > t0
        y1 = output[key][l][1]
        array[l][i+1]=y1
        if i==0:
            continue
        diff[l][i]=y1-y0


# Writing
if args.notext:
    print('Writing data to files...')
    f=open(job+prefix+'.dat','w')
    string = 'eV'
    for t in times:
        string = string + '\t' + t
    np.savetxt(f,array,delimiter='\t',header=string)
    f.close()

    f=open(job+'_diff.dat','w')
    string = 'eV'
    for i,t in enumerate(times):
        if i==0:
           continue
        string = string + '\t' + t + '-t0'
    np.savetxt(f,diff,delimiter='\t',header=string)
    f.close()

    print('Writing done.')
else:
    print('"--notext" flag')

# Plotting
fig,ax1=plt.subplots()
for i in range(1,len(diff[0])):
    plt.plot(diff[:,0],diff[:,i],'-',label='%s-%s'%(times[i],times[0]))
plt.legend(loc=2)
plt.xlabel("eV")
plt.ylabel("TA (arb. units)")
plt.axvline(1.94)
plt.title("TA, job %s/%s in folder %s"%(job,prefix,folder))
ax2 = ax1.twinx()
#plt.plot(array[:,0],array[:,1],'k-',label='BSE@t0')
plt.show()


print('Done.')



import sys
import numpy as np
from yambopy.units import ha2ev

def damp_it(ft, time, t_initial, damp_type="LORENTZIAN", damp_factor=0.1/ha2ev):
    ft_damped=np.empty_like(ft)

    if damp_type.upper() == "LORENTZIAN":
        ft_damped[:]=ft[:]*np.exp(-abs(time[:]-t_initial)*damp_factor)
    elif damp_type.upper() == "GAUSSIAN":
        ft_damped[:]=ft[:]*np.exp(-(time[:]-t_initial)**2*damp_factor**2)
    else:
        print("Wrong damping type ")
        sys.exit(0)

    return ft_damped



#
# Authors: MN
##
### Compute real space exction wavefunction when hole/electron is fixed.
import numpy as np
from yambopy.kpoints import build_ktree, find_kpt
from yambopy.tools.function_profiler import func_profile 
from tqdm import tqdm
import os

## Usage
"""
import numpy as np
from yambopy.dbs.excitondb import YamboExcitonDB
from yambopy.dbs.latticedb import YamboLatticeDB
from yambopy.dbs.wfdb import YamboWFDB
import os

iqpt = 1 # qpt index of exciton

# load lattice db
lattice = YamboLatticeDB.from_db_file(os.path.join('.', 'SAVE', 'ns.db1'))

# load exciton db
# note in case too many excitons, load only first few with `neigs' flag
# DO NOT forget to include all degenerate states when giving neigs flag !
#
filename = 'ndb.BS_diago_Q%d' % (iqpt)
excdb = YamboExcitonDB.from_db_file(lattice, filename=filename,
                                    folder=os.path.join('.', 'GW_BSE'),
                                    neigs = 20)

#Load the wavefunction database
wfdb = YamboWFDB(path='.', latdb=lattice,
                      bands_range=[np.min(excdb.table[:, 1]) - 1,
                      np.max(excdb.table[:, 2])])

## plot the exciton wavefunction with hole fixed at [0,0,0] (crystal coordinates)
# in a [1,1,1] supercell with 80 Ry wf cutoff
# I want to set the degeneracy threshold to 0.01 eV
# For example I want to plot the 3rd exciton, so iexe = 2 (python indexing )
#
excdb.real_wf_to_cube(iexe=2, wfdb=wfdb, fixed_position=[0,0,0], supercell=[1,1,1],
                        degen_tol=0.01, wfcCutoffRy=80, fix_particle='h')

## fix_particle = 'e' if you want to fix electron and plot hole density
## .cube will be dumped and use vesta to visualize it !
"""

##

@func_profile
def ex_wf2Real(Akcv, Qpt, wfcdb, bse_bnds, fixed_position,
               fix_particle='h', supercell=[1,1,1], wfcCutoffRy=-1,
               block_size=256):
    """
        Compute real-space exciton wavefunction when hole/electron is fixed.

        This is the main interface function that handles both resonant and anti-resonant parts
        of the exciton wavefunction in the case of non-TDA calculations.

    Args:
        Akcv (numpy.ndarray): Exciton wavefunction coefficients with shape:
                              - (Nstates,1,ns,k,c,v) for TDA
                              - (Nstates,2,ns,k,c,v) for non-TDA (2 for resonant/anti-resonant)
        Qpt (numpy.ndarray): Q-point of exciton in crystal coordinates
        wfcdb (YamboWFDB): Wavefunction database
        bse_bnds (list): Band range used in BSE [min_band, max_band] (python indexing)
        fixed_position (list): Position of fixed particle in crystal coordinates
        fix_particle (str): 'e' to fix electron, 'h' to fix hole (default)
        supercell (list): Supercell dimensions [nx,ny,nz]
        wfcCutoffRy (float): Wavefunction cutoff in Rydberg (-1 for no cutoff)
        block_size (int): Block size for memory-efficient computation.
        ## choosing lowe block_size will slight lower the memory requirment but also less faster

    Returns:
        tuple: (supercell_latvecs, atom_nums, atom_pos, exe_wfc_real)
               - supercell_latvecs: Supercell lattice vectors (3,3)
               - atom_nums: Atomic numbers.
               - atom_pos: Atomic positions in cartisian units
               - exe_wfc_real: Real-space exciton wavefunction (nstates, nspin, nspinor_electron, nspinor_hole,
                                                            Nx_grid, Ny_grid, Nz_grid)
                 note if nspin =2, then nspinor = 1, similarly, if nspinor = 2, nspin = 1.
    """
    ## first the resonat part
    supercell_latvecs,atom_nums,atom_pos,exe_wfc_real = \
            ex_wf2Real_kernel(Akcv[:,0], Qpt, wfcdb, bse_bnds, fixed_position,
                              fix_particle=fix_particle, supercell=supercell,
                              wfcCutoffRy=wfcCutoffRy, block_size=block_size,
                              ares=False, out_res=None)
    # for nonTDA add the anti-resonant part
    if Akcv.shape[1] == 2:
        supercell_latvecs,atom_nums,atom_pos,exe_wfc_real = \
            ex_wf2Real_kernel(Akcv[:,1], Qpt, wfcdb, bse_bnds, fixed_position,
            fix_particle=fix_particle, supercell=supercell,
            wfcCutoffRy=wfcCutoffRy, block_size=block_size,
                              ares=True, out_res=exe_wfc_real)

    return supercell_latvecs,atom_nums,atom_pos,exe_wfc_real



@func_profile
def ex_wf2Real_kernel(Akcv, Qpt, wfcdb, bse_bnds, fixed_position,
               fix_particle='h', supercell=[1,1,1], wfcCutoffRy=-1,
               block_size=256, ares=False, out_res=None):
    """
        Core kernel function for computing real-space exciton wavefunction.

        Computes either:
        - Akcv * psi_{kc}(r_e) * (psi_{k-Q,v}(r_h))^* (resonant part)
        - Akcv * psi_{kv}(r_e) * (psi_{k-Q,c}(r_h))^* (anti-resonant part)

        Note: For density, one must compute the absolute value squared.

    Args:
        Akcv (numpy.ndarray): Exciton coefficients [nstates,ns,nk,nc,nv]
        Qpt (numpy.ndarray): Q-point in crystal coordinates [3]
        wfcdb (YamboWFDB): Wavefunction database
        bse_bnds (list): BSE band range used in bse [nb1, nb2]. fortran indexing. i.e index starts from 1.
                        i.e nb1  and nb2 are same indices used in yambo input i.e
                        % BSEBands nb1 | nb2 % in yambo input
        fixed_position (list): Fixed particle position in crystal coords [3]
        fix_particle (str): 'e'=fix electron, 'h'=fix hole (default)
        supercell (list): Supercell dimensions [nx,ny,nz]
        wfcCutoffRy (float): Wavefunction cutoff in Rydberg (-1= full cutoff)
        block_size (int): Memory block size for computation. is a postive integer,
                        the default is 256 which is very good but uses more memory.
    # decrease it when you run into memory issues
        ares (bool): If True, compute anti-resonant part
        out_res (numpy.ndarray): Adds to this array
                                 and is returned instead of internally creating.
                                 Make sure it its consistant. (no internal checking done)

    Returns:
        tuple: (supercell_latvecs, atom_nums, atom_pos, exe_wfc_real)
        Natoms = (natom_in_unit_call * Nsupercell) + 1 (+1 due to hole/electron)
        - supercell_latvecs (numpy.ndarray): Supercell lattice vectors [3,3]
        - atom_nums (numpy.ndarray): Atomic numbers [Natoms]
        - atom_pos (numpy.ndarray): Atomic positions in cartisian units [Natoms,3]
        - exe_wfc_real (numpy.ndarray): Wavefunction in real space
          [nstates, nspin, nspinor_electron, nspinor_hole, FFTx, FFTy, FFTz]
          note if nspin =2, then nspinor = 1, similarly, if nspinor = 2, nspin = 1
    """
    #
    #
    if block_size < 1:
        print('Warning: Wrong block_size. setting to 1')
        block_size = 1
    #
    # Convert them to 
    for i in range(3):
        if supercell[i]%2 == 0:
            print('Warning : Even supercell given, so increasing'
                  + ' supercell size along %d direction by 1'%(i+1))
            supercell[i] = supercell[i] +1
    #
    #
    fix_particle = fix_particle.lower()
    bse_bnds = [min(bse_bnds)-1,max(bse_bnds)]
    assert bse_bnds[0] >= wfcdb.min_bnd, \
        "%d is used in bse but not found in wfcdb, load more wfcs" %(bse_bnds[0]+1)
    assert bse_bnds[1] <= wfcdb.min_bnd + wfcdb.nbands, \
        "%d is used in bse but not found in wfcdb, load more wfcs" %(wfcdb.min_bnd + wfcdb.nbands)
    bse_bnds = np.array(bse_bnds,dtype=int)-wfcdb.min_bnd
    
    # ns is number of collinear spin components
    nstates, ns, nk, nc, nv = Akcv.shape
    
    kpt_idx = wfcdb.ydb.kpoints_indexes
    sym_idx = wfcdb.ydb.symmetry_indexes
    nkBZ = len(sym_idx)
    assert nc + nv == bse_bnds[1]-bse_bnds[0], "Band mismatch"
    assert nk == nkBZ, "kpoint mismatch"
    #
    fixed_position = np.array(fixed_position)
    #
    hole_bnds = [bse_bnds[0],bse_bnds[0]+nv]
    elec_bnds = [bse_bnds[0]+nv,bse_bnds[1]]
    #
    if ares:
        # if anti-resonant part, we swap the electron and hole bands
        # first transpose c,v dimensions
        Akcv = Akcv.transpose(0,1,2,4,3)
        tmp  = hole_bnds
        hole_bnds = elec_bnds
        elec_bnds = tmp
        nstates, ns, nk, nc, nv = Akcv.shape

    lat_vec = wfcdb.ydb.lat.T
    blat = np.linalg.inv(lat_vec)
    gvecs_iBZ_idx = []
    fft_box = np.zeros(3,dtype=int)
    #
    
    for ik in range(len(wfcdb.gvecs)):
        idx_gvecs_tmp = np.arange(wfcdb.ngvecs[ik],dtype=int)
        if wfcCutoffRy > 0:
            tmp_gvecs = 2*np.pi*np.linalg.norm((wfcdb.gvecs[ik, :wfcdb.ngvecs[ik], :] 
                                        + wfcdb.kpts_iBZ[ik][None,:])@blat,axis=-1)
            idx_tmp = tmp_gvecs < np.sqrt(wfcCutoffRy)
            idx_gvecs_tmp = idx_gvecs_tmp[idx_tmp].copy()
        #
        gvecs_iBZ_idx.append(idx_gvecs_tmp)
        ## Get the fft box 
        min_fft_idx = np.min(wfcdb.gvecs[ik, :wfcdb.ngvecs[ik], :][idx_gvecs_tmp] , axis=0)
        max_fft_idx = np.max(wfcdb.gvecs[ik, :wfcdb.ngvecs[ik], :][idx_gvecs_tmp] , axis=0)
        assert np.min(max_fft_idx) >= 0 and np.max(min_fft_idx) < 0, "Invalid G-vectors"
        for i in range(3):
            fft_box[i] = max([fft_box[i], max_fft_idx[i] - min_fft_idx[i] + 3])
    
    # Compute nstates, nk, Nx, Ny, Nz object
    if out_res is None : print("Wfc FFT Grid : ",fft_box[0], fft_box[1], fft_box[2])
    #
    ##
    # find the nearest fft grid point.
    fixed_position = fixed_position.astype(np.float64)
    fx_pnt_int = np.floor(fixed_position)
    fixed_position -= fx_pnt_int
    fixed_position = np.round(fixed_position * fft_box) / fft_box
    fixed_position += fx_pnt_int
    # shift the position of hole to middle of supercell
    fixed_position += np.array(supercell)//2
    #
    if fix_particle == 'h':
        print("Position of the hole (reduced units) is set to : ",
              fixed_position[0], fixed_position[1], fixed_position[2])
    if fix_particle == 'e':
        print("Position of the electron (reduced units) is set to : ",
              fixed_position[0], fixed_position[1], fixed_position[2])
    #
    ktree = wfcdb.ktree #build_ktree(wfcdb.kBZ)
    #
    nspinorr = wfcdb.nspinor
    if out_res is not None:
        exe_wfc_real = out_res.reshape(nstates, ns, nspinorr, nspinorr,
                             supercell[0],fft_box[0],
                             supercell[1],fft_box[1],
                             supercell[2],fft_box[2])
    else:
        exe_wfc_real = np.zeros((nstates, ns, nspinorr, nspinorr,
                             supercell[0],fft_box[0],
                             supercell[1],fft_box[1],
                             supercell[2],fft_box[2]),
                            dtype=np.complex64)
    #
    Lx = np.arange(supercell[0],dtype=int)
    Ly = np.arange(supercell[1],dtype=int)
    Lz = np.arange(supercell[2],dtype=int)
    Lsupercells = np.zeros((supercell[0],supercell[1],supercell[2],3),dtype=int)
    Lsupercells[...,0], Lsupercells[...,1], Lsupercells[...,2] = np.meshgrid(Lx,Ly,Lz,indexing='ij')
    #
    FFTxx = np.fft.fftfreq(fft_box[0])
    FFTyy = np.fft.fftfreq(fft_box[1])
    FFTzz = np.fft.fftfreq(fft_box[2])
    FFTxx = FFTxx - np.floor(FFTxx)
    FFTyy = FFTyy - np.floor(FFTyy)
    FFTzz = FFTzz - np.floor(FFTzz)
    FFFboxs = np.zeros((fft_box[0],fft_box[1],fft_box[2],3))
    FFFboxs[...,0], FFFboxs[...,1], FFFboxs[...,2] = np.meshgrid(FFTxx,FFTyy,FFTzz,indexing='ij')
    #
    # ns is nspin
    exe_tmp_wf = np.zeros((nstates, ns, nspinorr, nspinorr, min(nk,block_size) ,
                           fft_box[0], fft_box[1], fft_box[2]),dtype=np.complex64)
    #
    exp_tmp_kL = np.zeros((min(nk,block_size) ,supercell[0],
                           supercell[1],supercell[2]),dtype=np.complex64)

    nblks = nk//block_size 
    nrem = nk%block_size
    if nrem > 0: nblks = nblks+1
    #
    pbar = tqdm(total=nk, desc="Ex-wf")
    #
    for ibk in range(nblks):
        ikstart = ibk*block_size
        ikstop = min(ikstart + block_size,nk)
        for ik in range(ikstart,ikstop):
            ## First get the electronic wfcs
            ik_ibz = kpt_idx[ik]
            isym = sym_idx[ik]
            wfc_tmp, gvecs_tmp = wfcdb.get_iBZ_wf(ik_ibz)
            wfc_tmp = wfc_tmp[:,elec_bnds[0]:elec_bnds[1],:,gvecs_iBZ_idx[ik_ibz]]
            gvecs_tmp = gvecs_tmp[gvecs_iBZ_idx[ik_ibz]]
            kvec = wfcdb.get_iBZ_kpt(ik_ibz)
            # get the rotated wf
            if isym != 0:
                sym_mat = wfcdb.ydb.sym_car[isym]
                time_rev = (isym >= len(wfcdb.ydb.sym_car
                                        ) / (1 + int(np.rint(wfcdb.ydb.time_rev))))
                kvec, wfc_tmp, gvecs_tmp = wfcdb.apply_symm(
                    kvec, wfc_tmp, gvecs_tmp, time_rev, sym_mat)

            kelec = kvec
            wfc_elec = wfc_tmp
            gvecs_elec = gvecs_tmp

            ## Do the same and get hole wfc
            ikhole = find_kpt(ktree, kelec-Qpt)
            ik_ibz = kpt_idx[ikhole]
            isym = sym_idx[ikhole]
            wfc_tmp, gvecs_tmp = wfcdb.get_iBZ_wf(ik_ibz)
            wfc_tmp = wfc_tmp[:,hole_bnds[0]:hole_bnds[1],:,gvecs_iBZ_idx[ik_ibz]]
            gvecs_tmp = gvecs_tmp[gvecs_iBZ_idx[ik_ibz]]
            kvec = wfcdb.get_iBZ_kpt(ik_ibz)
            # get the rotated wf
            if isym != 0:
                sym_mat = wfcdb.ydb.sym_car[isym]
                time_rev = (isym >= len(wfcdb.ydb.sym_car) / (1 + int(np.rint(wfcdb.ydb.time_rev))))
                kvec, wfc_tmp, gvecs_tmp = wfcdb.apply_symm(kvec, wfc_tmp, gvecs_tmp, time_rev, sym_mat)
            #
            khole = -kvec
            wfc_hole = wfc_tmp.conj()
            gvecs_hole = -gvecs_tmp

            if fix_particle == 'h':
                fx_kvec = khole
                fx_wfc = wfc_hole
                fx_gvec = gvecs_hole
                #
                ft_kvec = kelec
                ft_wfc = wfc_elec
                ft_gvec = gvecs_elec
            else :
                ft_kvec = khole
                ft_wfc = wfc_hole
                ft_gvec = gvecs_hole
                #
                fx_kvec = kelec
                fx_wfc = wfc_elec
                fx_gvec = gvecs_elec
            # compute
            ## NM : Donot perform FFT as we only need it for one point.
            exp_fx = np.exp(2*np.pi*1j*((fx_gvec + fx_kvec[None,:])@fixed_position))
            fx_wfc *= exp_fx[None,None,None,:]
            fx_wfc = np.sum(fx_wfc,axis=-1) #(spin,bnd,spinor)
            ns1, nbndc, nspinorr, ng = ft_wfc.shape
            #if ft_ikpt not in prev_ikpts:
            ft_wfcr = wfcdb.to_real_space(ft_wfc.reshape(-1,nspinorr,ng),ft_gvec, grid=fft_box)
            ft_wfcr = ft_wfcr.reshape(ns1,nbndc,nspinorr,fft_box[0],fft_box[1],fft_box[2])
            exp_kx_r = np.exp(2*np.pi*1j*FFFboxs.reshape(-1,3)@ft_kvec).reshape(FFFboxs.shape[:3])
            ft_wfcr *= exp_kx_r[None,None,None,...]
            #
            fx_wfc = fx_wfc.astype(np.complex64)
            ft_wfcr = ft_wfcr.astype(np.complex64)
            #
            if fix_particle == 'h':
                np.einsum('nscv,svy,scxijk->nsxyijk',Akcv[:,:,ik,...].astype(np.complex64),fx_wfc,ft_wfcr,
                          optimize=True,out=exe_tmp_wf[:,:,:,:,ik-ikstart])
            else :
                np.einsum('nscv,scx,svyijk->nsxyijk',Akcv[:,:,ik,...].astype(np.complex64),fx_wfc,ft_wfcr,
                          optimize=True,out=exe_tmp_wf[:,:,:,:,ik-ikstart])
            #
            #exe_tmp_wf[:,:,:,ik-ikstart] *= exp_kx_r[...].reshape(FFFboxs.shape[:3])[None,None,None]
            exp_tmp_kL[ik-ikstart] = np.exp(1j*2*np.pi*np.einsum('...x,x->...',Lsupercells,ft_kvec))
            #
            # update progess bar
            pbar.update(1)
            #
        ## perform gemm operation 
        total_gemms_t = nstates*ns*nspinorr**2
        exp_tmp_kL_tmp = exp_tmp_kL.reshape(len(exp_tmp_kL),-1)[:(ikstop-ikstart)].T
        exe_tmp_wf_tmp = exe_tmp_wf.reshape(nstates,ns,nspinorr,nspinorr,-1,np.prod(fft_box))
        exe_tmp_wf_tmp = exe_tmp_wf_tmp[...,:(ikstop-ikstart),:]
        #
        for igemms in range(total_gemms_t):
            ii, iis, jj, kk = np.unravel_index(igemms, (nstates,ns,nspinorr,nspinorr))
            Ctmp = (exp_tmp_kL_tmp @ exe_tmp_wf_tmp[ii, iis, jj, kk ])
            Ctmp = Ctmp.reshape(supercell[0], supercell[1], supercell[2],
                                fft_box[0], fft_box[1], fft_box[2])
            Ctmp *= (1.0/np.prod(supercell))
            exe_wfc_real[ii, iis, jj, kk ] += Ctmp.transpose(0,3,1,4,2,5)
    #
    exe_wfc_real   = exe_wfc_real.reshape(nstates,ns,nspinorr,nspinorr,
                                            supercell[0]*fft_box[0],
                                            supercell[1]*fft_box[1],
                                            supercell[2]*fft_box[2])
    #
    # compute postioon of fixed particle in cart units 
    fixed_position_cc = lat_vec@fixed_position
    Lsupercells = Lsupercells.reshape(-1,3)#/np.array(supercell)[None,:]
    Lsupercells = Lsupercells@lat_vec.T
    atom_pos = Lsupercells[:,None,:] + wfcdb.ydb.car_atomic_positions[None,:,:]
    atom_pos = np.append(atom_pos.reshape(-1,3),fixed_position_cc[None,:],axis=0)
    supercell_latvecs = lat_vec*np.array(supercell)[None,:]
    ## Make atomic numbers 
    atom_nums = np.zeros(len(atom_pos),dtype=wfcdb.ydb.atomic_numbers.dtype)
    attmp = atom_nums[:-1].reshape(-1,len(wfcdb.ydb.atomic_numbers))
    attmp[...]= wfcdb.ydb.atomic_numbers[None,:]
    atom_nums[-1] = 200
    #
    return supercell_latvecs,atom_nums,atom_pos,exe_wfc_real

#def compute_exc_wfc_real(path='.', bse_dir='SAVE', iqpt=1, nstates=[1],
#                          fixed_position=[0,0,0], fix_particle='h', aveg=True, supercell=[1,1,1],
#                         wfcCutoffRy=-1, phase=False, block_size=256):
#    #
#    lattice = YamboLatticeDB.from_db_file(os.path.join(path, 'SAVE', 'ns.db1'))
#    filename = 'ndb.BS_diago_Q%d' % (iqpt)
#    excdb = YamboExcitonDB.from_db_file(lattice, filename=filename,
#                                                 folder=os.path.join(path, bse_dir),
#                                                 Load_WF=True, neigs=max(nstates))
#    # Load the wavefunction database
#    wfdb = YamboWFDB(path=path, latdb=lattice,
#                      bands_range=[np.min(excdb.table[:, 1]) - 1,
#                        np.max(excdb.table[:, 2])])
#    #
#    Akcv = excdb.get_Akcv()[min(nstates)-1:max(nstates)]
#    excQpt = excdb.car_qpoint
#    #
#    # Convert the q-point to crystal coordinates
#    Qpt = wfdb.ydb.lat @ excQpt

#    sc_latvecs, atom_nums, atom_pos, real_wfc = ex_wf2Real(Akcv, Qpt, wfdb, [np.min(excdb.table[:, 1]),
#                np.max(excdb.table[:, 2])], fixed_position=fixed_position,
#                          fix_particle=fix_particle, supercell=supercell, 
#                          wfcCutoffRy=wfcCutoffRy, block_size=block_size)
#    #
#    #
#    nstates_range = np.arange(nstates[0],nstates[1],dtype=int)
#    density = np.abs(real_wfc)**2

#    if fix_particle == 'h': name_file = 'electron'
#    else: name_file = 'hole'
#    if real_wfc.shape[1] != 1:
#        print("phase plot only works for nspin = 1 and nspinor == 1")
#        phase = False
#    if phase: 
#        phase = np.sign(real_wfc.real) #np.sign(np.angle(real_wfc))
#        density *= phase
#    if aveg:
#        real_wfc = np.sum(density,axis=(0,1,2))
#        real_wfc = real_wfc/np.max(np.abs(real_wfc))
#        write_cube('exe_wf_avg_%s_%d-%d.cube' %(name_file,nstates[0],nstates[1]),
#                   real_wfc, sc_latvecs, atom_pos, atom_nums, header='Real space exciton wavefunction')
#    else:
#        real_wfc = np.sum(density,axis=(1,2))
#        for i in range(len(real_wfc)):
#            real_wfc1 = real_wfc[i]/np.max(np.abs(real_wfc[i]))
#            write_cube('exe_wf_%s_%d.cube' %(name_file,nstates[i]), real_wfc1, sc_latvecs,
#                       atom_pos, atom_nums, header='Real space exciton wavefunction')



##if __name__ == "__main__":




import numpy as np
import sys

def Fourier_Interpolation(ft, fw, time,freqs,mode="T2W"):
    #
    # I assume constant time-step and regular frequency grid
    # otherwise this subroutine needs to be generalized
    #
    t_step=time[1]-time[0]
    f_step=freqs[1]-freqs[0]
    #
    if mode.upper() == "T2W":
        fw[:,:]=0.0+0.0j
        for i_w in range(fw.shape[1]):
            for i_c in range(ft.shape[0]):
                    fw[i_c,i_w] = fw[i_c,i_w]+np.sum(ft[i_c,:]*np.exp(1j*freqs[i_w]*time[:]))*t_step
    elif mode.upper() == "W2T":
        ft[:,:]=0.0+0.0j
        for i_w,i_t in zip(range(ft.shape[1]),range(fw.shape[1])):
            for i_c in range(ft.shape[0]):
                ft[i_c,i_t] = ft[i_c,i_t]+fw[i_c,i_w]*np.exp(-1j*freqs[i_w]*time[i_t])*f_step





#
# Authors: MN
#
import argparse
import os
from yambopy.dbs.excitondb import YamboExcitonDB
from yambopy.dbs.latticedb import YamboLatticeDB
#
def run_exc_sort(args):
    parser = argparse.ArgumentParser(description="Write the sorted energies and intensities to a file.")
    #
    parser.add_argument("--save", type=str, default=".", help="Path to the SAVE directory (location of ns.db1, as in PATH/SAVE). Default current directory.")
    parser.add_argument("-J", "--jobdir", type=str, default="SAVE", metavar="DIR", help="BSE job directory (as in PATH/DIR). Default SAVE")
    parser.add_argument("--iqpt", type=int, default=1, help="Q-point index. Default 1")
    args = parser.parse_args(args)

    nsfile = os.path.join(args.save, "ns.db1")
    lattice = YamboLatticeDB.from_db_file(nsfile)
    #
    filename = f"ndb.BS_diago_Q{args.iqpt}"
    excdb = YamboExcitonDB.from_db_file(lattice, filename=filename,
                                        folder=args.jobdir,
                                        Load_WF=False, neigs=-1)
    excdb.write_sorted(prefix=f"o-{args.jobdir.split('/')[-1]}.exc_qpt{args.iqpt}_sorted")




# This script checks convergence for all k-points

from yambopy import *
import numpy as np
import matplotlib.pyplot as plt

# pack files of convergence GW calculations
pack_files_in_folder('gw_conv')

# Start Analyser
ya = YamboAnalyser('gw_conv')

# Plot of all the k-points converging one parameter
ya.plot_gw_all_kpoints_convergence(tag='EXX')

ya.plot_gw_all_kpoints_convergence(tag='Bnds')

ya.plot_gw_all_kpoints_convergence(tag='NGsBlk')

ya.plot_gw_all_kpoints_convergence(tag='GbndRnge')



from yambopy import *
import numpy as np
import matplotlib.pyplot as plt
from netCDF4 import Dataset
import os
from yambopy.plot.plotting import BZ_Wigner_Seitz
from yambopy.kpoints import expand_kpoints
from yambopy.units import ha2ev, ev2cm1

def phonon_overlap(mode1,mode2):
    """
    Compute the overlap between phonon eigenvectors e_ka,n (q).
    This can be useful to disentangle phonon modes along a q-direction.

    - See Zacharias and Giustino PRR 2020, Eq. (55)

    k -> atomic index
    a -> Cartesian index
    n -> branch index
    q -> momentum index

    (the eigenvectors are complex)

    :: mode1 is e (q1)
    :: mode2 is e (q2)
    
    Eigenvectors are read from prefix.dynq files (QE)

    The overlap matrix is:

    M_nm(q1,q2) = \\sum_ka e_ka,n (q1) e^*_ka,m (q2)
    """
    
    # Overlap matrix
    M = np.einsum('nka,mka->nm',mode1,np.conj(mode2))

    # Debug (explicit equivalent of einsum)
    #M2 = np.zeros([nmodes,nmodes],dtype=np.complex)
    #for n in range(nmodes):
    #    for m in range(nmodes):
    #        for k in range(natoms):
    #            for a in range(3):
    #                M2[n,m] += mode1[n,k,a]*np.conj(mode2[m,k,a])

    return M    

class YamboRefineElphDB():
    """
    This class:
    
    (i) replaces the phonon energies stored in the ndb.elph_gkkp_expanded
    databases with those resulting from matdyn.x, which include ASR and LOTO
    proper treatment

    (ii) for 2D systems, replaces the values of g_LO(q=0), which is zero in a QE
    calculation, with the average absolute value of the closest q/=0 elements, 
    approximating the correct behaviour (cusp) for dense enough q-sampling

    - Inputs: 
    :: YamboLatticeDB object
    :: folder of ndb.elph_gkkp_expanded databases
    :: path of prefix.bn file (from matdyn.x) including filename
    :: [optional] index of LO branch to correct at q=0 
    :: [optional] path of prefix.dyn files (from matdyn.x) including filename w/out q-label
    :: REPLACE: if True, will prompt the user about editing the databases

    - Usage and main variables:
        :: yelph = YamboRefineElphDB(ylat, ... args ... )

        :: Has all relevant attributes for easy testing
        :: Contains testing functions to plot q-BZ map and phonon dispersion to be manually edited

        :: Uses branch disentangling to check for nearest-neighbour LO modes. This can be done
           by reading eigenvectors either from yambo or from prefix.dyn files.
           BEWARE!! Orderings may change in the two cases!!
    """ 
    def __init__(self,lattice,filename='ndb.elph_gkkp_expanded',folder_gkkp='SAVE',out_matdyn='prefix.freq',LO_ind=None,matdyn_filename='prefix.dyn',REPLACE=False):
        # Find correct database names
        filename = "%s/%s"%(folder_gkkp,filename)
        self.frag_filename = filename + "_fragment_"
        self.matdyn = out_matdyn        
        self.dyn_files = matdyn_filename
        if LO_ind is not None: self.LO = LO_ind

        # Check if databases exist. Exit if header is missing.
        try: database = Dataset(filename)
        except: raise FileNotFoundError("error opening header %s"%filename)
        try: 
            database_frag = Dataset("%s1"%self.frag_filename)
            database_frag.close()
        except: 
            raise FileNotFoundError("Database fragment at q=0 not detected")
    
        #read qpoints
        self.alat    = lattice.alat
        self.rlat    = lattice.rlat
        self.sym_car = lattice.sym_car
        self.qpoints = database.variables['PH_Q'][:].T
        self.car_qpoints = np.array([ q/self.alat for q in self.qpoints ])
        #read dimensions of electron phonon parameters
        self.nmodes, self.nqpoints, self.nkpoints, self.nbands = database.variables['PARS'][:4].astype(int)
        self.nmodes, self.nqpoints, self.nkpoints, b_1, b_2 = database.variables['PARS'][:5].astype(int)
        if b_1>b_2: # Old database (no GkkpBands in PARS)
            self.nbands = b_1
            self.b_in, self.b_out = [0,self.nbands-1]
        else: # New database (PARS with GkkpBands)
            self.b_in, self.b_out = [b_1-1,b_2-1]
            self.nbands = b_2-b_1+1
        self.natoms = int(self.nmodes/3)
        try: # Check if K-point list is provided (upon expansion)
            self.kpoints_elph = database.variables['PH_K'][:].T
            self.car_kpoints = np.array([ k/self.alat for k in self.kpoints_elph ])
            database.close()
        except KeyError:
            database.close()

        #Check how many databases are present
        self.nfrags = self.nqpoints
        for iq in range(self.nqpoints):
            if not os.path.isfile("%s%d"%(self.frag_filename,iq+1)):
                self.nfrags = iq
                break
        if self.nfrags!=self.nqpoints: raise FileNotFoundError("Found: %d fragments. Needed: %d"%(self.nfrags,self.nqpoints))

        #Check presence of matdyn output
        if not os.path.isfile(self.matdyn): raise FileNotFoundError("Matdyn 'prefix.freq' file not found")
        # Now the checks are done, we can continue
        # Frequencies part 
        self.read_frequencies_yambo()
        self.read_matdyn()
        self.expand_kpoints()
        self.expand_frequencies()
        # LO mode part
        if LO_ind is not None:
            self.find_nearest_qneighbor()
            self.find_nearest_modes(mode='yambo')
            self.get_LO_modes()

        # Now that we have everything, we can edit the DBs
        if REPLACE:
            self.replace_frequencies()
            self.replace_LO_modes()

    def read_frequencies_yambo(self):
        """
        Read phonon frequencies in eV
        """
        self.ph_energies  = np.zeros([self.nfrags,self.nmodes])

        for iq in range(self.nfrags):
            fil = self.frag_filename + "%d"%(iq+1)
            database = Dataset(fil,'r')
            self.ph_energies[iq] = np.sqrt(database.variables['PH_FREQS%d'%(iq+1)][:])*ha2ev
            database.close()

    def read_eigenvectors(self,iq,mode='matdyn'):
        """ Wrapper for either matdyn or yambo reading 
        """
        if mode=='matdyn': return self.read_eigenvectors_matdyn(iq)
        if mode=='yambo':  return self.read_eigenvectors_yambo(iq)
        

    def read_eigenvectors_yambo(self,iq):
        """
        Read phonon eigenvectors
        """
        ph_eigenvectors = np.zeros([self.nmodes,self.natoms,3],dtype=np.complex64)

        fil = self.frag_filename + "%d"%(iq+1)
        database= Dataset(fil)
        eigs_q = database.variables['POLARIZATION_VECTORS'][:].T
        ph_eigenvectors = eigs_q[0,:,:,:] + eigs_q[1,:,:,:]*1j
        database.close()
        
        return ph_eigenvectors

    def read_matdyn(self):
        """ Read prefix.freq file
            - Read qpoints (ibz)
            - Read frequencies (ibz)
            - Transform into bz
        """
        invcm2eV = 1.23981e-4 
        f = open(self.matdyn,'r')
        lines = f.readlines()
        Nq_ibz = int(lines[0].split()[-2])
        lines = lines[1:]
        Nlines = len(lines)
        if Nlines % Nq_ibz != 0: raise ValueError("Something wrong in matdyn file!")
        nstep = int(Nlines/Nq_ibz)
        qpts_matdyn = []
        freqs_matdyn = []
        for iq in range(Nq_ibz):
            qpts_matdyn.append([float(i) for i in lines[iq*nstep].split()])
            aux = []
            for i1 in range(1,nstep): aux += [ float(i) for i in lines[iq*nstep+i1].split() ]
            freqs_matdyn.append(aux)
        self.qpts_matdyn  = np.array(qpts_matdyn)/self.alat[0]
        self.freqs_matdyn = np.array(freqs_matdyn)*invcm2eV
        if np.any(self.freqs_matdyn<0.): print("[WARNING] NEGATIVE frequencies found in the matdyn file!")

    def expand_kpoints(self,atol=1e-6,verbose=1):
        """
        Expand matdyn qpoints
        """

        weights, kpoints_indexes, symmetry_indexes, kpoints_full = expand_kpoints(self.qpts_matdyn,self.sym_car,self.rlat,atol=atol)

        if verbose: print("%d kpoints expanded to %d"%(len(self.car_kpoints),len(kpoints_full)))

        #set the variables
        self.weights_ibz        = weights
        self.qpoints_indices    = kpoints_indexes
        self.symmetry_indices   = symmetry_indexes
        self.iku_matdyn_qpoints = np.array([k*self.alat for k in kpoints_full])
        self.car_matdyn_qpoints = kpoints_full

    def expand_frequencies(self):
        self.matdyn_ph_energies = np.zeros([self.nqpoints,self.nmodes])
        for iq in range(self.nqpoints): self.matdyn_ph_energies[iq] = self.freqs_matdyn[self.qpoints_indices[iq]]
    
    def read_eigenvectors_matdyn(self,iq):
        """
        Read eigenvectors from dyn files
        """
        ph_raw_energies = np.zeros(self.nmodes)
        ph_eigen        = np.zeros([self.nmodes,self.natoms,3],dtype=np.complex)
        f = open(self.dyn_files+str(iq+1), "r")
        start_line = -(self.natoms+1)*self.nmodes-1
        end_line = -1
        data = f.readlines()[start_line:end_line]
        # First read frequencies (convert to eV, so far unused)
        im=0
        for line in data:
            if 'freq' in line:
                ph_raw_energies[im] = float(line.split()[-2])/ev2cm1
                im+=1
        # Next read eigenmodes
        im=0
        for im in range(self.nmodes):
            for iat in range(self.natoms):
                ind = im*(self.natoms+1)+1+iat
                tmp = data[ind].split()
                ph_eigen[im,iat,0] = float(tmp[1])+1j*float(tmp[2])
                ph_eigen[im,iat,1] = float(tmp[3])+1j*float(tmp[4])
                ph_eigen[im,iat,2] = float(tmp[5])+1j*float(tmp[6])

        return ph_eigen 

    def find_nearest_qneighbor(self, max_NN=12, atol=1e-4):
        """ Finds the first nearest-neighbour points
            to q=(0,0,0) in the q-grid. This is intended to be
            the first shell of NN in a *regular grid*. 

            Function can be improved to find first NN across x,y,z if the
            grid is anisotropic.
            
            :: max_NN must be >= number of first NN
        """

        from sklearn.neighbors import KDTree
        
        qpts = self.car_qpoints

        # Build a KDTree from the grid points
        tree = KDTree(qpts)

        # Find the indices of the max_NN nearest neighbors of G
        distances, indices = tree.query(np.array([qpts[0]]), k=10)
        distances, indices = [distances[0][1:],indices[0][1:]]

        # Select only the first NN up to atol
        min_dist = min(distances)
        q_NN = []
        for iq, qdist in enumerate(distances):  
            if np.isclose(qdist,min_dist,atol=atol,rtol=atol): q_NN.append(indices[iq])

        self.q_NN = sorted(q_NN)
        self.N_NN = len(self.q_NN)

    def find_nearest_modes(self,mode='matdyn',verbose=1):
        """
        Find the indices of the LO-adjacent mode at q/=0
        corresponding to the first NN of q=0.

        We use the overlap matrix.
        """
        # q=0 modes
        eig0 = self.read_eigenvectors(0,mode=mode)
        # NN modes
        self.modes_NN = []
        for iq_bz in self.q_NN:
            iq_ibz = self.qpoints_indices[iq_bz]
            # (i) Read all the modes at iq_NN (in the IBZ)
            eigq = self.read_eigenvectors(iq_ibz,mode=mode)
            # (ii) Get the overlap of all the modes with those at q=0
            Overlap_matrix = phonon_overlap(eig0,eigq)
            # (iii) Find the mode with highest overlap with LO mode
            overlap = 0.
            best_mode = self.LO
            for im in range(self.nmodes):
                ref_value = np.abs(Overlap_matrix[self.LO,im])
                if ref_value > overlap:
                    overlap   = ref_value
                    best_mode = im
            # (iv) Obtain mode index of NN of LO modes
            self.modes_NN.append(best_mode)
            if verbose:
                if best_mode==self.LO:
                    print("==== No crossings detected ===")
                    print("Phonon mode #%d at q 0 remains phonon mode #%d at q_ibz #%d / q_bz #%d"%(self.LO,best_mode,iq_ibz,iq_bz))
                else:
                    print("==== Crossings detected ===")
                    print("Phonon mode #%d at q 0 becomes phonon mode #%d at q_ibz #%d / q_bz $%d"%(self.LO,best_mode,iq_ibz,iq_bz))

    def get_LO_modes(self):
        """ Obtain values for LO modes
        """
        # gkkp[q][k][bnd2][bnd1][mode][cmplx]
        dvscf_NN = np.zeros([self.N_NN,self.nkpoints,self.nbands,self.nbands,self.nmodes],dtype=np.complex64)

        for iq,Q in enumerate(self.q_NN):
            fil = self.frag_filename + "%d"%(Q+1)
            database = Dataset(fil)
            dvscf_aux = database.variables['ELPH_GKKP_Q%d'%(Q+1)][:]
            dvscf_NN[iq] = dvscf_aux[:,:,:,:,0] + 1j*dvscf_aux[:,:,:,:,1]
            database.close()

        # Check integrity of elph values
        if np.isnan(dvscf_NN).any(): print('[WARNING] NaN values detected in elph database.')
        
        # Compute modulus of NN g_LO(q_NN) and average it to g_LO_new
        g_LO_new = np.zeros([self.nkpoints,self.nbands,self.nbands])
        for iq in range(self.N_NN): g_LO_new += np.abs(dvscf_NN[iq,:,:,:,self.modes_NN[iq]])
        self.g_LO_new = g_LO_new/len(self.q_NN)    
        
        # Read dvscf at q=0    
        fil = self.frag_filename + "1"
        database = Dataset(fil)
        self.dvscf_0 = database.variables['ELPH_GKKP_Q1'][:]
        database.close()
        self.g_LO_old = self.dvscf_0[:,:,:,self.LO,0] + 1j*self.dvscf_0[:,:,:,self.LO,1]
        
    def replace_frequencies(self):
        """ 
        """
        usr_inp = input("Smooth frequencies from matdyn.x will replace raw ones in the ndb.elph_gkkp* databases. Proceed ['y','n']? ")
        if usr_inp != 'y':
            print("Frequencies not replaced.")
        else:
        
            aux_ph_energies = (self.matdyn_ph_energies/ha2ev)**2. 
        
            for iq in range(self.nfrags):
                fil = self.frag_filename + "%d"%(iq+1)
                database = Dataset(fil,'r+')
                database.variables['PH_FREQS%d'%(iq+1)][:] = aux_ph_energies[iq]
                database.close()
            
            print("Frequencies replaced.")

    def replace_LO_modes(self):
        """
        """
        usr_inp = input("The LO matrix elements at q=0 will be replaced with the |average| of their q nearest neighbors. Proceed ['y','n']? ")
        if usr_inp != 'y':
            print("LO matrix elements not replaced.")
        else:

            # Replace values for LO mode
            self.dvscf_0[:,:,:,self.LO,0] = self.g_LO_new
            self.dvscf_0[:,:,:,self.LO,1] = 0.
            
            # Write dvscf at q=0
            fil = self.frag_filename + "1"
            database = Dataset(fil,"r+")
            database.variables['ELPH_GKKP_Q1'][:] = self.dvscf_0
            database.close()        
                
            print("LO phonon modes replaced.")
        
    ### MATERIAL-DEPENDENT TEST FUNCTIONS ###
    def plot_qBZ(self,title="",save=False,mode='yambo'):
        """Plot annotated q-points in qBZ
        """
        if mode=='yambo':  points = self.car_qpoints
        if mode=='matdyn': points = self.car_matdyn_qpoints
        fig = plt.figure(figsize=(9,9))
        ax = plt.gca()
        ax.add_patch(BZ_Wigner_Seitz(self.rlat,color='black',linewidth=1.))
        ax.set_aspect('equal')
        ax.set_title(title)
        ax.scatter(points[:,0],points[:,1],marker='H',s=80,color='teal',linewidth=0.5,edgecolors='black',label='expanded')
        for i_k,kpt in enumerate(points):
            kx,ky = kpt[0],kpt[1]
            ax.annotate(i_k, (kx,ky), color='red', xytext=(kx+0.0005,ky+0.001), fontsize=5)
        if save: plt.savefig('%s.pdf'%title)
        plt.show()

    def find_path(self,indx_0,indx_1,band_indx,path='GMKG',mode='yambo'):
        """
        path: only GMKG

        :: indx_0 = list of high-symmetry point indices (e.g. G,M,K)
        :: indx_1 = list of high-symm. indx. in band path including end point
        :: band_indices = indices of band path 

        The above variables must be taken manually from plot_qBZ maps.

        Example for 12x12x1 grid:

        indx_0 = [0,31,142]
        indx_1 = [0,6,8,12]
        band_indx = [0,1,7,13,19,25,31,118,142,124,88,34,0] 
       
        """
        if path!='GMKG': raise ValueError("Only 'GMKG' path option is possible")

        if mode=='yambo': 
            pts=self.car_qpoints
            energies=self.ph_energies
        if mode=='matdyn':
            pts=self.car_matdyn_qpoints
            energies=self.matdyn_ph_energies

        iG, iM, iK   = indx_0
        Sindx        = indx_1
        band_indices = band_indx
        pts[iM] = np.abs(pts[iM])

        Np = len(band_indices)
        Nb = energies.shape[1]
        band_momenta = np.zeros(Np)
        band_energies= np.zeros([Np,Nb])
        nrm = np.linalg.norm
        for ip in range(1,Np): band_momenta[ip]  = band_momenta[ip-1]+nrm(pts[band_indices[ip]]-pts[band_indices[ip-1]])
        for ip in range(Np):   band_energies[ip] = energies[band_indices[ip],:] 

        return Sindx, band_momenta, band_energies

    def plot_bands(self,bnd_rng,indx_0,indx_1,band_indx,plt_show=True):
        """
        Phonon bands from gkkp (1) and matdyn (2)
        bnd_rng: [b_first, b_last]

        See above function for meaning of indx arguments.
        """
        Sindx, x1, y1 = self.find_path(indx_0,indx_1,band_indx)
        _,     x2, y2 = self.find_path(indx_0,indx_1,band_indx,mode='matdyn') 
        y1 = y1*1000.
        y2 = y2*1000.

        title='Phonon dispersion'
        clr1='teal'
        clr2='orange'
        flnm='Phon_Bands.pdf'
        ms1=10
        ms2=6
        lab1 = 'ndb.elph'
        lab2 = 'matdyn'

        fig = plt.figure(figsize=(8,6))
        ax = plt.gca()
        b_i, b_o = bnd_rng
        ax.set_xlim(x1[0]-0.01,x1[-1]+0.01)

        Spts = [x1[Sindx[0]],x1[Sindx[1]],x1[Sindx[2]],x1[Sindx[-1]]]
        Slabs = ['G','M','K','G']

        for PT in Spts: ax.axvline(PT,c='black')
        ax.set_xticks(Spts)
        ax.set_xticklabels(Slabs)
        ax.set_ylabel('Energy (meV)')
        for ib in range(b_i,b_o+1):
            if ib==b_i:
                ax.plot(x1,y1[:,ib],'.-',markersize=ms1,c=clr1,label=lab1)
                ax.plot(x2,y2[:,ib],'^-',markersize=ms2,c=clr2,label=lab2)
            else:
                ax.plot(x1,y1[:,ib],'.-',markersize=ms1,c=clr1)
                ax.plot(x2,y2[:,ib],'^-',markersize=ms2,c=clr2)

        plt.legend()
        plt.savefig(flnm)
        if plt_show: plt.show()

    ### END MATERIAL-DEPENDENT TEST FUNCTIONS ###



from yambopy import YamboLatticeDB,YamboQPDB # Load yambo netcdf databases
from qepy import Path                        # Define path in k-space
import numpy as np
import matplotlib.pyplot as plt
from math import sqrt

# Define path in reduced coordinates using Class Path
npoints = 10
path = Path([ [[  0.0,  0.0,  0.0],'$\Gamma$'],
              [[  0.5,  0.0,  0.0],'M'],
              [[1./3.,1./3.,  0.0],'K'],
              [[  0.0,  0.0,  0.0],'$\Gamma$']], [int(npoints*2),int(npoints),int(sqrt(5)*npoints)] )

# Read Lattice information from SAVE
## Note: we do not expand the kpts because QP database is in the IBZ
lat  = YamboLatticeDB.from_db_file(filename='SAVE/ns.db1',Expand=False)
# Read QP database
ydb  = YamboQPDB.from_db(filename='ndb.QP',folder='qp-gw')
n_top_vb = 3 # Top valence band index starting from 0
fermie = np.max(ydb.eigenvalues_qp[:,n_top_vb]) # Energy shift to top valence

# 1. Find scissor operator for valence and conduction bands

fig = plt.figure(figsize=(6,4))
ax  = fig.add_axes( [ 0.20, 0.20, 0.70, 0.70 ])
ax.set_xlabel('$E_{KS}$')
ax.set_ylabel('$E_{GW}$')

ydb.plot_scissor_ax(ax,n_top_vb+1)

plt.show()

# 2. Plot of KS and QP eigenvalues NOT interpolated along the path
ks_bs_0, qp_bs_0 = ydb.get_bs_path(lat,path)

fig = plt.figure(figsize=(4,5))
ax = fig.add_axes( [ 0.20, 0.20, 0.70, 0.70 ])

ks_bs_0.plot_ax(ax,legend=True,c_bands='r',label='KS')
qp_bs_0.plot_ax(ax,legend=True,c_bands='b',fermie=fermie,label='QP-GW')

plt.show()

# 3. Interpolation of KS and QP eigenvalues

ks_bs, qp_bs = ydb.interpolate(lat,path,what='QP+KS',lpratio=20)

fig = plt.figure(figsize=(4,5))
ax = fig.add_axes( [ 0.20, 0.20, 0.70, 0.70 ])

ks_bs.plot_ax(ax,legend=True,c_bands='r',label='KS')
qp_bs.plot_ax(ax,legend=True,c_bands='b',fermie=fermie,label='QP-GW')

plt.show()

# 4. Comparison of not-interpolated and  interpolated eigenvalues

fig = plt.figure(figsize=(4,5))
ax = fig.add_axes( [ 0.20, 0.20, 0.70, 0.70 ])

ks_bs_0.plot_ax(ax,legend=True,c_bands='r',label='KS')
qp_bs_0.plot_ax(ax,legend=True,c_bands='b',fermie=fermie,label='QP-GW')
ks_bs.plot_ax(ax,legend=True,c_bands='g',label='KS int.')
qp_bs.plot_ax(ax,legend=True,c_bands='k',fermie=fermie,label='QP-GW int.')

plt.show()



#
# License-Identifier: GPL
#
# Copyright (C) 2024 The Yambo Team
#
# Authors: FP, RR
#
# This file is part of the yambopy project
#
import os
from glob import glob
from qepy.lattice import Path
from yambopy import *
from yambopy.units import *
from yambopy.plot.plotting import add_fig_kwargs,BZ_Wigner_Seitz
from yambopy.lattice import replicate_red_kmesh, calculate_distances, car_red
from yambopy.kpoints import get_path
from yambopy.tools.funcs import gaussian, lorentzian

class ExcitonDispersion():
    """
    Class to obtain exciton information at all momenta

    - Dispersion plot (under development)
    - Plots of exciton weights in q-space

    :: Lattice is an instance of YamboLatticeDB
    :: nexcitons is the number of excitonic states - by default it is taken from the Q=1 database

    NB: so far does not support spin-polarised exciton plots (should be implemented when needed!)
    NB: only supports BSEBands option in yambo bse input, not BSEEhEny
    """

    def __init__(self,lattice,nexcitons=None,folder='.'):

        if not isinstance(lattice,YamboLatticeDB):
            raise ValueError('Invalid type for lattice argument. It must be YamboLatticeDB')
        
        files   = glob(folder+'/ndb.BS_diago_Q*')
        nqpoints  = len(files)

        # Check
        if not nqpoints==lattice.ibz_nkpoints :
            raise ValueError("Incomplete list of qpoints (%d/%d)"%(nqpoints,lattice.ibz_nkpoints)) 
    
        dbs_are_consistent, spin_is_there = self.db_check(lattice,nqpoints,folder)
        if nexcitons is None: nexcitons = self.ntransitions

        # Read
        car_qpoints         = np.zeros((nqpoints,3))
        exc_energies        = np.zeros((nqpoints,nexcitons))
        exc_eigenvectors    = np.zeros((nqpoints,nexcitons,self.ntransitions),dtype=complex)
        exc_tables          = np.zeros((nqpoints,self.ntransitions,5),dtype=int)
        for iQ in range(nqpoints):
            exc_obj = YamboExcitonDB.from_db_file(lattice,filename=folder+'/ndb.BS_diago_Q%d'%(iQ+1))
            if iQ==0: car_qpoints[iQ] = np.array([0.,0.,0.])
            else:     car_qpoints[iQ] = exc_obj.car_qpoint
            exc_energies[iQ,:]        = exc_obj.eigenvalues[:nexcitons].real
            exc_eigenvectors[iQ,:]    = exc_obj.eigenvectors[:nexcitons]    
            exc_tables[iQ,:]          = exc_obj.table

        # Set up variables
        self.nqpoints     = nqpoints
        self.nexcitons    = nexcitons
        self.car_qpoints  = car_qpoints
        self.red_qpoints  = car_red(car_qpoints,lattice.rlat)
        self.lattice      = lattice
        self.exc_energies = exc_energies
        self.exc_tables   = exc_tables

        # Reshape eigenvectors if possible
        if dbs_are_consistent and not spin_is_there: self.exc_eigenvectors = self.reshape_eigenvectors(exc_eigenvectors)
        else:                                        self.exc_eigenvectors = exc_eigenvectors

        # Necessary lattice information
        self.alat = lattice.alat
        self.rlat = lattice.rlat

    def db_check(self,lattice,nqpoints,folder):
        """
        Check nexcitons and ntransitions in each database
        """
        nexcitons_each_Q    = np.zeros(nqpoints,dtype=int)
        for iQ in range(nqpoints):
            exc_obj = YamboExcitonDB.from_db_file(lattice,filename=folder+'/ndb.BS_diago_Q%d'%(iQ+1))
            nexcitons_each_Q[iQ] = exc_obj.nexcitons
            if iQ==0: tbl = exc_obj.table

        is_spin_pol = len(np.unique(tbl[:,3]))>1 or len(np.unique(tbl[:,4]))>1
        is_consistent = np.all(nexcitons_each_Q==nexcitons_each_Q[0])
        if not is_consistent: 
            print("[WARNING] BSE Hamiltonian has different dimensions for some Q.")
            print("          Taking the minimum number of transitions, be careful.")
            self.ntransitions = np.min(nexcitons_each_Q)
        else:
            if is_spin_pol: print("[WARNING] Spin-polarised excitons, only partially supported")
            self.ntransitions = nexcitons_each_Q[0]
            valence_bands    = np.unique(tbl[:,1]) - 1
            conduction_bands = np.unique(tbl[:,2]) - 1
            self.nkpoints    = np.max(tbl[:,0])
            self.nvalence    = len(valence_bands)
            self.nconduction = len(conduction_bands)

        return is_consistent, is_spin_pol

    def reshape_eigenvectors(self,eigenvectors):
        """
        eigenvectors in:  [nqpoints, nexcitons, ntransitions]
        eigenvectors out: [nqpoints, nexcitons, nkpoints, nvalence, nconduction]

        TODO: Extend to spin-polarised case
        """
        nq, nexc, nk, nv, nc = self.nqpoints, self.nexcitons, self.nkpoints, self.nvalence, self.nconduction
        reshaped_eigenvectors = np.zeros((nq,nexc,nk,nv,nc),dtype=complex)
        #print(eigenvectors[2,5,2])
        for iQ in range(nq):
            for i_exc in range(nexc): 
                reshaped_eigenvectors[iQ,i_exc,:,:,:] = eigenvectors[iQ,i_exc,:].reshape([nk,nv,nc])
        #print(reshaped_eigenvectors[2,5,0,1,0])    

        return reshaped_eigenvectors

    @add_fig_kwargs
    def plot_Aweights(self,data,plt_show=False,plt_cbar=False,**kwargs):
        """
        2D scatterplot in the q-BZ of the quantity A_{iq}(iexc,ik,ic,iv).

        Any real quantity which is a function of only the q-grid may be supplied.
        The indices iq,inu,ib1,ib2 are user-specified.

        - if plt_show plot is shown
        - if plt_cbar colorbar is shown
        - kwargs example: marker='H', s=300, cmap='viridis', etc.

        NB: So far requires a 2D system.
            Can be improved to plot BZ planes at constant k_z for 3D systems.
        """

        qpts = self.car_qpoints

        # Input check
        if len(data)!=len(qpts):
            raise ValueError('Something wrong in data dimensions (%d data vs %d qpts)'%(len(data),len(qpts)))

        # Global plot stuff
        self.fig, self.ax = plt.subplots(1, 1)
        if self.nqpoints<self.nkpoints:  c_BZ_borders='black'
        if self.nqpoints==self.nkpoints: c_BZ_borders='white'
        self.ax.add_patch(BZ_Wigner_Seitz(self.lattice,color=c_BZ_borders))


        if plt_cbar:
            if 'cmap' in kwargs.keys(): color_map = plt.get_cmap(kwargs['cmap'])
            else:                       color_map = plt.get_cmap('viridis')
        lim = 1.05*np.linalg.norm(self.rlat[0])
        self.ax.set_xlim(-lim,lim)
        self.ax.set_ylim(-lim,lim)

        # Reproduce plot also in adjacent BZs
        BZs = shifted_grids_2D(qpts,self.rlat)
        for qpts_s in BZs: plot=self.ax.scatter(qpts_s[:,0],qpts_s[:,1],c=data,**kwargs)

        if plt_cbar: self.fig.colorbar(plot)

        plt.gca().set_aspect('equal')

        if plt_show: plt.show()
        else: print("Plot ready.\nYou can customise adding savefig, title, labels, text, show, etc...")

    #####################################
    # Dispersion plot under development #
    #####################################
    def get_dispersion(self, path):
        """ 
        Obtain dispersion along symmetry lines.
        
        Similar to band plots in k-space, check YamboExcitonDB for more comments

        :: path is instance of Path class
        """
        qpoints = self.red_qpoints
        qpath    = np.array(path.kpoints)

        rep = list(range(-1,2))
        qpoints_rep, qpoints_idx_rep = replicate_red_kmesh(qpoints,repx=rep,repy=rep,repz=rep)
        exc_indexes = get_path(qpoints_rep,qpath)[1] #indices are second output
        exc_qpoints  = np.array(qpoints_rep[exc_indexes])
        exc_indexes = qpoints_idx_rep[exc_indexes]
        
        # Here assuming same ordering in index expansion between k-yambopy and q-yambo...
        energies = self.exc_energies[self.lattice.kpoints_indexes]
        energies_path  = energies[exc_indexes]
        
        ybs_disp = YambopyBandStructure(energies_path, exc_qpoints, kpath=path)
        return ybs_disp
        
    def get_dispersion_interpolated(self):
        """ Interpolated with SkwInterpolator
        """
    
    def plot_exciton_disp_ax(self,ax,path,**kwargs):
        ybs_disp = self.get_dispersion(path)
        print(ybs_disp.nbands)
        print(ybs_disp.nkpoints)
        print(ybs_disp._xlim)
        print(ybs_disp.nbands)
        return ybs_disp.plot_ax(ax) 

    @add_fig_kwargs
    def plot_exciton_disp(self,path,**kwargs):
        import matplotlib.pyplot as plt
        fig = plt.figure()
        ax = fig.add_subplot(1,1,1)
        self.plot_exciton_disp_ax(ax,path)
        return fig
    
    def plot_dispersion():
        """ Do plot
        """
    
    def __str__(self):
        lines = []; app = lines.append
        app(" Exciton Dispersion ")
        app(" Number of qpoints:                    %d"%self.nqpoints)
        app(" Number of exciton branches read:      %d"%self.nexcitons)
        app(" Total number of excitons/transitions: %d"%self.ntransitions)
        return "\n".join(lines)



# Copyright (C) 2018 Henrique Pereira Coutada Miranda
# All rights reserved.
#
# This file is part of yambopy
#
import unittest
import os
from yambopy.dbs.qpdb import YamboQPDB
test_path = os.path.join(os.path.dirname(__file__),'..','..','data','refs','gw')

class TestYamboQPDB(unittest.TestCase):

    def test_yamboqpdb(self):

        #open qpdb
        qpdb = YamboQPDB.from_db(folder=test_path)

        #get qp dbb
        eigenvalues_qp, eingenvalues_dft, lifetimes, z = qpdb.get_qps()

        #plot bs
        qpdb.plot_bs(show=False)

        print(qpdb)

if __name__ == '__main__':
    unittest.main()



# Copyright (c) 2018, Alejandro Molina-Sanchez
# All rights reserved.
#
# This file is part of the yambopy project
#
from yambopy import *
from yambopy.plot  import *
import os

class YamboRTMovie(object):
    """
    Create a file with data of real time simulations performed with Yambo
    """
    def __init__(self,job_string,path='.'):
        """
        Parameters:
            job_string - the job_string used for yambo. yambo -J <job_string>
            path       - the folder where the yambo run was made
        """

        #look for the save folder
        self.save=path+'/SAVE'
        if not os.path.isdir(self.save):
            raise ValueError('SAVE folder not found in %s'%self.save)

        # FP: changed to new electron class, to be tested
        YamboElectronsDB.__init__(self,save=self.save)

        self.job_string = job_string
        self.data = {"lattice": self.lat,
                     "atypes": self.atomic_numbers,
                     "atoms": self.atomic_positions}

        self.atoms = None
        self.excitons = None
 
        #use YamboOut to read the polarization-time
        self.path = path

        #try to find o-* files in path, if not use path/job_string
        paths = [path, "%s/%s"%(path,job_string)]
        for p in paths:
            y = YamboOut(p,save_folder=path)
            polarization = y.get_data(('polarization'))
            carriers     = y.get_data(('carriers')) 
            external     = y.get_data(('external')) 
            #if we read the files then continue
            if polarization != {}:
                break

        #trap the errors here
        if polarization == {}:
            raise ValueError('Could not find the o-*polarization files in %s'%paths)
        if carriers == {}:
            raise ValueError('Could not find the o-*carriers files in %s'%paths)

        #we just use one of them
        key = list(polarization)[0]
        for key,value in polarization[key].items():
            self.data[key] = value
        key = list(carriers)[0]
        for key,value in carriers[key].items():
            self.data[key] = value
        key = list(external)[0]
        for key,value in external[key].items():
            self.data[key] = value

    def get_excitons(self,min_intensity=0.1,max_energy=4,Degen_Step=0.0):
        filename = "%s/o-%s.exc_E_sorted"%(self.path,self.job_string)
        if not os.path.isfile(filename):
            os.system("cd %s; ypp -e s -J %s"%(self.path,self.job_string))
        self.excitons = np.loadtxt(filename)

        #filter with energy
        self.excitons = self.excitons[self.excitons[:,0]<max_energy]

        #filter with intensity
        self.excitons = self.excitons[self.excitons[:,1]>min_intensity]

        #filter with degen
        if Degen_Step:

            #create a list with differences in energy
            new_excitons = []
            prev_exc = 0
            for exc in self.excitons:
                e,i,index = exc
                #if the energy of this exciton is too diferent then we add it to the list
                if abs(e-prev_exc)<Degen_Step:
                    new_excitons[-1][1] += i
                    continue
                new_excitons.append([e,i,index])
                intensity = 0
                prev_exc = e
            self.excitons = np.array(new_excitons)

        #create dictionary with excitons 
        excitons = self.data["excitons"]
        for e,intensity,i in self.excitons:
            exciton = {"energy": e,
                       "intensity": intensity,
                       "index": i}
            excitons.append(exciton)
        return self.excitons

    def get_wavefunctions(self, FFTGvecs=30,
                          Cells=[1,1,1], Hole=[0,0,0],
                          Direction="123", Format="x",
                          Degen_Step=0.0100,
                          MinWeight=1e-8,
                          repx=range(-1,2), repy=range(-1,2), repz=range(-1,2),
                          wf=False):
        """
        Collect all the wavefuncitons with an intensity larger than self.threshold
        Parameters:
          FFTGvecs   - Number of FFTGvecs. Related to how accurate the representation is
          Cells      - Number of cells to plot in real space
          Hole       - Define the hole position in cartesian coordinates
          Direction  - Choose the directions to plot along
          Format     - Choose the format to plot in. Can be: x for xcrysden or g for gnuplot (default: 'x' for xcrysden)
          Degen_Step - Threshold to merge degenerate states. If the difference in energy between the states is smaller than
                       this value their wavefunctions will be plotted together
          repx       - Number or repetitions along the x direction
          repy       - Number or repetitions along the x direction
          repz       - Number or repetitions along the x direction
          wf         - Get the wavefuncitons in real space or not (default: False)
        """
        if self.excitons is None:
            raise ValueError( "Excitons not present. Run YamboBSEAbsorptionSpectra.get_excitons() first" )
        self.data["excitons"] = []

        #create a ypp file using YamboIn for reading the wavefunction
        yppwf = YamboIn('ypp -e w -V all',filename='ypp.in',folder=self.path)
        yppwf['Format'] = Format
        yppwf['Direction'] = Direction
        yppwf['FFTGvecs'] = [FFTGvecs,'Ry']
        yppwf['Degen_Step'] = [Degen_Step,'eV']
        yppwf['Hole'] = [Hole,'']
        yppwf['Cells'] = [Cells,'']

        #create a ypp file using YamboIn for reading the excitonic weights
        yppew = YamboIn('ypp -e a',filename='ypp.in',folder=self.path)
        yppew['MinWeight'] = MinWeight
        yppew['Degen_Step'] = Degen_Step

        keywords = ["lattice", "atoms", "atypes", "nx", "ny", "nz"]
        for exciton in self.excitons:
            #get info
            e,intensity,i = exciton

            if wf:
                ##############################################################
                # Excitonic Wavefunction
                ##############################################################
                #create ypp input for the wavefunction file and run
                yppwf["States"] = "%d - %d"%(i,i)
                yppwf.write("%s/yppwf_%d.in"%(self.path,i))

                filename = "o-%s.exc_%dd_%d%s"%(self.job_string,len(Direction),i,{"g":"","x":".xsf"}[Format] )
                if not os.path.isfile(filename):
                    os.system("cd %s; ypp -F yppwf_%d.in -J %s"%(self.path,i,self.job_string))

                #read the excitonic wavefunction
                if Format == 'x':
                    ewf = YamboExcitonWaveFunctionXSF()
                else:
                    ewf = YamboExcitonWaveFunctionGnuplot()
                ewf.read_file("%s/%s"%(self.path,filename))
                data = ewf.get_data()
                for word in keywords:
                    if word in data:
                        self.data[word] = data[word]

                #calculate center of mass of atoms
                lat = np.array(data["lattice"])
                center_atom = np.zeros([3])
                for atype,x,y,z in data["atoms"]:
                    center_atom += np.array([x,y,z])
                center_atom /= len(data["atoms"])
                center_atom_red = car_red([center_atom],lat)[0]

                #shift wavefunctions grid to center of mass
                nx = data['nx'] 
                ny = data['ny'] 
                nz = data['nz']

                #make center_atom_red commensurate with fft
                center_atom_red = center_atom_red * np.array([nx,ny,nz])
                center_atom_red_int = [int(x) for x in center_atom_red]
                displacement = np.array([nx,ny,nz])/2-center_atom_red_int
                dx,dy,dz = displacement

                # shift grid
                # http://www.xcrysden.org/doc/XSF.html 
                dg  = np.array(data["datagrid"]).reshape([nz,ny,nx])
                dg  = np.roll(dg,dx,axis=2)
                dg  = np.roll(dg,dy,axis=1)
                dg  = np.roll(dg,dz,axis=0)
                data["datagrid"] = dg.flatten()

                #shift atoms
                atoms = []
                dx,dy,dz = red_car([displacement/np.array([nx,ny,nz],dtype=float)],lat)[0]
                for atype,x,y,z in data["atoms"]:
                    atoms.append([atype,x+dx,y+dy,z+dz])
                self.data["atoms"] = atoms

            ##############################################################
            # Excitonic Amplitudes
            ##############################################################
            #create ypp input for the amplitudes file and run
            yppew["States"] = "%d - %d"%(i,i)
            yppew.write("%s/yppew_%d.in"%(self.path,i))

            filename = "%s/o-%s.exc_weights_at_%d"%(self.path,self.job_string,i)
            if not os.path.isfile(filename):
                os.system("cd %s; ypp -F yppew_%d.in -J %s"%(self.path,i,self.job_string))

            #read the excitonic weigths
            ew = YamboExcitonWeight(filename,save=self.save,path=self.path)
            qpts, weights = ew.calc_kpts_weights(repx=repx,repy=repy,repz=repz)

            ############
            # Save data
            ############
            exciton = {"energy": e,
                       "intensity": intensity,
                       "weights": weights,
                       "qpts": qpts,
                       "index": i}
            if wf:
                exciton["hole"] = Hole
                exciton["datagrid"] = np.array(data["datagrid"])

            self.data["excitons"].append(exciton)

    def write_json(self,filename="dynamics"):
        """ Write a jsonfile with the absorption spectra and the wavefunctions of certain excitons
        """
        JsonDumper(self.data,"%s.json"%filename)



#
# This file is part of yambopy
#
import xml.etree.ElementTree as ET
import numpy as np

"""
List of functions to read xml data from .xml files
"""

def get_xml_attrib(xml_tree,tag,attrib,repeated=False):
    """ 
    Extract xml attribute
    
    If repeated = True, first occurrence of attribute is stored.
    Else, its last occurrence is stored.
    """
    attribute =  None
    for element in xml_tree.iter(tag=tag):
        attribute = element.attrib[attrib]
        if repeated: break

    if attribute is None: raise ValueError('xml attribute %s not found'%attrib)
    else: return attribute
    
def get_xml_data(xml_tree,tag,as_type=float,repeated=False):
    """ 
    Extract xml data as specified type
    
    If repeated = True, first occurrence of attribute is stored.
    Else, its last occurrence is stored.
    """    
    data = None
    for element in xml_tree.iter(tag=tag):
        data_as_str = element.text.split()
        if len(data_as_str)>1: data = [as_type(d_str) for d_str in data_as_str]
        else: data = as_type(data_as_str[0])
        if repeated: break
        
    if data is None: raise ValueError('xml data within tag %s not found'%tag)
    else: return data

def get_xml_nk_bands(xml_tree):
    """
    Function to specifically get kpoint (cartesian) coordinates and corresponding eigenvalues (in Hartree)
    """    
    k_points_car  = []
    k_eigenvalues = []
    k_occupations = []
    for ks_energies in xml_tree.iter(tag='ks_energies'):
        k_points_car.append(  get_xml_data(ks_energies,'k_point',as_type=float)     )
        k_eigenvalues.append( get_xml_data(ks_energies,'eigenvalues',as_type=float) )
        k_occupations.append( get_xml_data(ks_energies,'occupations',as_type=float) )

    k_points_car  = np.array(k_points_car)
    k_eigenvalues = np.array(k_eigenvalues)
    k_occupations = np.array(k_occupations)

    return k_points_car, k_eigenvalues, k_occupations



from __future__ import print_function, division
#
# Author: Henrique Pereira Coutada Miranda
# Check the convergence of the coulomb cutoff for a BSE calculation using yambo
#
from yambopy import *
from qepy import *
from yambocommandline import *
from schedulerpy import *
from functools import partial
import matplotlib.pyplot as plt
import multiprocessing
import argparse
import sys

prefix = "bn"
yambo = "yambo"
p2y = 'p2y'
pw = 'pw.x'
layer_separations = [10,15,20,25,30,35,40]
scf_kpoints  = [ 9, 9,1]
nscf_kpoints = [12,12,1]
nbands = 20
ecutwf = 50

scheduler = Scheduler.factory

# create the quantum espresso input file
def get_inputfile():
    """ Define a Quantum espresso input file for boron nitride
    """ 
    qe = PwIn()
    qe.set_atoms([['N',[ 0.0, 0.0,0.5]],
                ['B',[1./3,2./3,0.5]]])
    qe.atypes = {'B': [10.811, "B.pbe-mt_fhi.UPF"],
                 'N': [14.0067,"N.pbe-mt_fhi.UPF"]}

    qe.control['prefix'] = "'%s'"%prefix
    qe.control['wf_collect'] = '.true.'
    qe.control['pseudo_dir'] = "'../../../pseudos/'"
    qe.system['celldm(1)'] = 4.7
    qe.system['celldm(3)'] = 14/qe.system['celldm(1)']
    qe.system['ecutwfc'] = ecutwf
    qe.system['occupations'] = "'fixed'"
    qe.system['nat'] = 2
    qe.system['ntyp'] = 2
    qe.system['ibrav'] = 4
    qe.kpoints = scf_kpoints
    qe.electrons['conv_thr'] = 1e-10
    return qe

#run the self consistent calculation
def scf(layer_separation,folder='scf'):
    if not os.path.isdir(folder):
        os.makedirs(folder)
    qe = get_inputfile()
    qe.system['celldm(3)'] = layer_separation/qe.system['celldm(1)']
    qe.control['calculation'] = "'scf'"
    qe.write('%s/%s.scf'%(folder,prefix))

#run the non-self consistent calculation
def nscf(layer_separation,folder='nscf'):
    if not os.path.isdir(folder):
        os.makedirs(folder)
    qe = get_inputfile()
    qe.control['calculation'] = "'nscf'"
    qe.electrons['diago_full_acc'] = ".true."
    qe.electrons['conv_thr'] = 1e-8
    qe.system['nbnd'] = nbands
    qe.system['force_symmorphic'] = ".true."
    qe.system['celldm(3)'] = layer_separation/qe.system['celldm(1)']
    qe.kpoints = nscf_kpoints
    qe.write('%s/%s.nscf'%(folder,prefix))
    
def database(shell,output_folder,nscf_folder='nscf'):
    if not os.path.isdir('%s/SAVE'%output_folder):
        print('preparing yambo database...')
        shell.add_command('mkdir -p %s'%nscf_folder)
        shell.add_command('cd %s/%s.save; %s; %s; cd ../../../../'%(nscf_folder,prefix,p2y,yambo))
        shell.add_command('mv %s/%s.save/SAVE %s'%(nscf_folder,prefix,output_folder))
        print('done!')

def run_job(layer_separation,nthreads=1,work_folder='bse_cutoff',cut=False):
    """
    Given a layer separation run the calculation
    1. scf calculation with QE
    2. nscf calculation
    3. BSE with yambo
    """
    #check if the calculation exists
    done_stamp = '%s/%d/done'%(work_folder,layer_separation)
    print(done_stamp)
    if os.path.isfile(done_stamp):
        return

    print("layer separation: %d bohr    cutoff:"%layer_separation, cut)
    root_folder = "%s/%d"%(work_folder,layer_separation)
    shell = scheduler()
    if not os.path.isdir(root_folder):
        shell.add_command( 'mkdir -p %s'%root_folder )

    # 1. run the ground state calculation
    print("scf cycle")
    print("kpoints",scf_kpoints)
    scf(layer_separation,folder="%s/scf"%root_folder)
    shell.add_command("cd %s/scf; mpirun -np %d %s -inp %s.scf > scf.log "%(root_folder,nthreads,pw,prefix))
    shell.add_command("cd ../../../")

    # 2. run the non self consistent calculation
    print("nscf cycle")
    print("kpoints",nscf_kpoints)
    src ='%s/scf/%s.save'%(root_folder,prefix)
    dst ='%s/nscf/%s.save'%(root_folder,prefix)
    nscf(layer_separation,folder="%s/nscf"%root_folder)

    shell.add_command('cp -r %s %s'%(src,dst) )
    shell.add_command("cd %s/nscf; mpirun -np %d %s -inp %s.nscf > nscf.log"%(root_folder,nthreads,pw,prefix))
    shell.add_command("cd ../../../" )

    # generate the database
    database(shell,'%s'%root_folder,nscf_folder="%s/nscf"%root_folder)
    shell.run()
    #wait for execution

    # 3. calculate the absorption spectra
    y = YamboIn.from_runlevel('%s -r -X s -o b -k sex -y d -V all'%yambo,executable=yambo,folder=root_folder)

    if cut:
        y['CUTGeo'] = 'box z'
        y['CUTBox'] = [0,0,layer_separation-2]

        y['RandQpts'] = 1000000
        y['RandGvec'] = [1,'Ry']

    y['FFTGvecs'] = [20,'Ry']
    y['NGsBlkXs'] = [1,'Ry']          #local field effects
    y['BndsRnXs'] = [1,nbands]            #number of bands for static screening

    y['KfnQP_E']  = [2.91355133,1.0,1.0] #scissor operator
    y['BSEBands'] = [4,5]                #number of bands in BSE kernel
    y['BEnRange'] = [[4.0,8.0],'eV']     #energy range to plot optical absorption
    y['BEnSteps'] = 500                  #energy steps in the range
    y.write('%s/yambo_run.in'%root_folder)

    shell = scheduler()
    shell.add_command('cd %s; mpirun -np %d %s -F yambo_run.in -J %d'%(root_folder,nthreads,yambo,layer_separation))
    shell.add_command('touch done')
    shell.run()

def run(mpthreads=1,nthreads=1,work_folder='bse_cutoff',cut=True):

    if (mpthreads > 1):
        p = multiprocessing.Pool(nthreads)
        run = partial(run_job,nthreads=nthreads,work_folder=work_folder,cut=cut)
        try:
            #reversed list because of load imbalance
            p.map(run, reversed(layer_separations))
        except KeyboardInterrupt:
            print("Caught KeyboardInterrupt, terminating workers")
            p.terminate()
            p.join()

    else:
        for layer_separation in layer_separations:
            run_job(layer_separation,nthreads=nthreads,work_folder=work_folder,cut=cut)

def plot(work_folder,filename,cut):
    ax = plt.gca()
    for layer_separation in layer_separations:
        root_folder = "%s/%d"%(work_folder,layer_separation)
    
        #gather the results
        pack_files_in_folder(root_folder)

    #plot the results
    ya = YamboAnalyser(work_folder)
    print(ya)
    ax = ya.plot_bse(('eps_q1'),cols=(2,),ax=ax)

    if cut: title = "with coulomb cutoff"
    else:   title = "without coulomb cutoff"

    plt.title(title)
    if filename is None: filename = "%s.pdf"%work_folder
    plt.savefig(filename)
    plt.show()

if __name__ == "__main__":

    #parse options
    parser = argparse.ArgumentParser(description='Convergence test of the colomb cutoff')
    parser.add_argument('-r' ,'--run',      action="store_true", help='Run the calculation')
    parser.add_argument('-c' ,'--cut',      action="store_true", help='Use coulomb cutoff')
    parser.add_argument('-p' ,'--plot',     action="store_true", help='Run the analysis')
    parser.add_argument('-f' ,'--plotfile', help='name of the plot file', default=None)
    parser.add_argument('-t' ,'--nthreads', help='threads for yambo', default=1, type=int)
    parser.add_argument('-mp' ,'--mpthreads', help='theads using python multiprocessing module', default=1, type=int)
    args = parser.parse_args()
    print("yambo using %d threads"%args.nthreads)
    print("multiprocessing using %d threads"%args.mpthreads)

    if len(sys.argv)==1:
        parser.print_help()
        sys.exit(1)

    cut = args.cut
    
    #choose work_folder
    if cut: 
        work_folder = "bse_cutoff_cut"
    else:
        work_folder = "bse_cutoff"

    if args.run:
        run(args.mpthreads,args.nthreads,work_folder,cut)
    if args.plot:
        plot(work_folder,args.plotfile,cut)



import time
from schedulerpy import *

"""
This file contains the basic functions needed to check and manage workflows

    - wait_for_job: Let the python execution sleep until job completion
    - TODO: submit_job, check_for_job_completion, ... 
"""

def wait_for_job(shell,run_dir,time_step=10.):
    """
    Let the python execution sleep until job completion.
    
    - shell: schedulerpy object relative to submitted job
    - run_dir: directory where the job is being run
    - time_step: checking period (seconds)
    
    Scheduler types supported:
    
        - bash
        - slurm
    """
    job_status = shell.check_job_status(run_dir)
    condition = job_status=='R' or job_status=='PD' or job_status=='CG'
    while condition:
        time.sleep(time_step)
        job_status = shell.check_job_status(run_dir) 
        condition = job_status=='R' or job_status=='PD' or job_status=='CG'
        
def wait_for_all_jobs(shell_list,run_dir_list,time_step=10.):
    """
    As above, but waits for completion of a list of jobs
    """
    if len(shell_list) != len(run_dir_list):
        raise UserWarning('ERROR in parallel job management: list of job ids not corresponding to list run directories.')
    
    # Get initial statuses
    job_status_list = []
    conditions = []
    for i in range(len(shell_list)):
        job_status =  shell_list[i].check_job_status(run_dir_list[i])
        job_status_list.append( job_status )
        condition = job_status=='R' or job_status=='PD' or job_status=='CG'
        conditions.append(condition)
    
    # Dynamically evaluate all conditions
    while any(conditions):
        time.sleep(time_step)
        job_status_list = []
        conditions = []
        for i in range(len(shell_list)):
            job_status =  shell_list[i].check_job_status(run_dir_list[i])
            job_status_list.append( job_status )
            condition = job_status=='R' or job_status=='PD' or job_status=='CG'
            conditions.append(condition)
        
        # Remove finished calculation (i.e., False) from checks
        for i, condition in enumerate(conditions):
            if not condition: del shell_list[i]

def wait_for_setup_operations(filename,run_dir,time_step=10.):
    """
    To be used inside submission scripts which (i) have depedencies, (ii) need setup.

    Example: yambo calculation can only run after nscf is finished (dependency), but
             needs the SAVE folder to be created by the python master process as well (setup)

    When the dependencies are fulfilled, the job is submitted by the scheduler. This function delays
    the execution of the actual code (i.e., yambo in the example above) until the setup is also completed
    by the master python process. 

    This usually lasts a much shorter time than the actual execution. By executing this function inside the
    submission script we avoid hanging the workflow.

    - filename: an empty file (i.e., named 'SETUP_DONE') touched by the setup routine to signal end of setup.
                when filenames 
    """
    stderr = True
    while stderr:
        time.sleep(time_step)
        p = subprocess.Popen(['ls','%s'%filename],stdout=subprocess.PIPE,stderr=subprocess.PIPE,cwd=run_dir)
        stdout,stderr = p.communicate() 
    
    



# Copyright (c) 2023-2025, Claudio Attaccalite,
#                          Myrta Gruening, Mao Yunchen
# All rights reserved.
#
# This file is part of the yambopy project
# Calculate linear response from real-time calculations (yambo_nl)
# Modified to allow generalisation
#
import numpy as np
from yambopy.units import ha2ev,fs2aut
from yambopy.nl.external_efield import Divide_by_the_Field
from tqdm import tqdm
import sys
import os
from abc import ABC,abstractmethod
from yambopy.nl.nl_analysis import Xn_from_signal
#
#
# Derived class for monochromatic signal
#    
class Xn_from_sine(Xn_from_signal):
        def set_defaults(self):
            EFIELDS = ["SIN","SOFTSIN"]
            if self.efield["name"] not in EFIELDS:
                raise ValueError(f"Invalid electric field for frequency mixing analysis. Expected one of: {EFIELDS}")
            for i_n in range(len(self.pumps)):
                if self.pumps[i_n]["name"] != 'none':
                    raise ValueError("This analysis is for one monochromatic field only.")
            if self.solver == '':
                self.solver = 'full'
            self.out_dim = self.X_order + 1
            if self.samp_mod== "":
                self.samp_mod="linear"
            return

        def set_sampling(self,ifrq):
            samp_order = 2*self.X_order + 1
            if self.nsamp == -1:
                self.nsamp = samp_order
            T_period = 2.0 * np.pi / self.freqs[ifrq]
            T_range, out_of_bounds = self.update_time_range(T_period)
            if (out_of_bounds):
                print(f'User range redifined for frequency {self.freqs[ifrq]* ha2ev:.3e} [eV]')
            print(f"Time range: {T_range[0] / fs2aut:.3f} - {T_range[1] / fs2aut:.3f} [fs]")
            return T_range
        
        def update_time_range(self,T_period): # not sure if this is a general or specific method - let it here for the moment
            T_range = self.T_urange
            out_of_bounds = False
            if T_range[0] <= 0.0:
                T_range[0] = self.time[-1] - T_period
            if T_range[1] > 0.0:
                T_range[1] = T_range[0] + T_period
            else:
                T_range[1] = self.time[-1]                
            if T_range[1] > self.time[-1]:
                T_range[1] = slef.time[-1]
                T_range[0] = T_range[1] - T_period
                out_of_bound = True
            return T_range, out_of_bounds

        def define_matrix(self,T_i,ifrq):
            M_size = len(T_i)
            M = np.zeros((M_size, M_size), dtype=np.cdouble)
            M[:, 0] = 1.0
            W = self.freqs[ifrq]
            for i_n in range(1, self.X_order+1):
                exp_neg = np.exp(-1j * i_n* W * T_i, dtype=np.cdouble)
                exp_pos = np.exp(1j * i_n * W * T_i, dtype=np.cdouble)
                M[:, i_n] = exp_neg
                M[:, i_n +self.X_order] = exp_pos
            return M

        def output_analysis(self,out,to_file=True):
            for i_order in range(self.X_order + 1):
                T = 10000.0
                for i_f in range(self.n_runs):
                    out[i_order, i_f, :] *= Divide_by_the_Field(self.efields[i_f], i_order)
                    T_period = 2.0 * np.pi / self.freqs[i_f]
                    Trange, _ = self.update_time_range(T_period)
                    T = min(T,Trange[0])
                out[i_order,:,:]*=self.get_Unit_of_Measure(i_order)
                run_info = self.append_runinfo(T)
                if (to_file):
                    output_file = f'o{self.prefix}.YamboPy-X_probe_order_{i_order}'
                    header = "E[eV] " + " ".join([f"X{i_order}/Im({d}) X{i_order}/Re({d})" for d in ('x','y','z')])
                    if self.l_out_current:
                        output_file = f'o{self.prefix}.YamboPy-Sigma_probe_order_{i_order}'
                        header = "E[eV] " + " ".join([f"S{i_order}/Im({d}) S{i_order}/Re({d})" for d in ('x','y','z')])
                    values = np.column_stack((self.freqs * ha2ev, out[i_order, :, 0].imag, out[i_order, :, 0].real,
                                      out[i_order, :, 1].imag, out[i_order, :, 1].real,
                                      out[i_order, :, 2].imag, out[i_order, :, 2].real))
                    np.savetxt(output_file, values, header=header, delimiter=' ', footer="Harmonic analysis results")
                    outf = open(output_file, "a")
                    outf.writelines(run_info)
                    outf.close()
                else:
                    print(run_info)
                    return (self.freqs, out)

        def reconstruct_signal(self,out,to_file=True):
            Seff = np.zeros((self.n_runs, 3, len(self.time)), dtype=np.cdouble)
            for i_f in tqdm(range(self.n_runs)):
                for i_d in range(3):
                    for i_order in range(self.X_order + 1):
                        freq_term = np.exp(-1j * i_order * self.freqs[i_f] * self.time)
                        Seff[i_f, i_d, :] += out[i_order, i_f, i_d] * freq_term
                        Seff[i_f, i_d, :] += np.conj(out[i_order, i_f, i_d]) * np.conj(freq_term)    
            for i_f in tqdm(range(self.n_runs)):
                values = np.column_stack((self.time / fs2aut, Seff[i_f, 0, :].real, Seff[i_f, 1, :].real, Seff[i_f, 2, :].real))
                output_file = f'o{self.prefix}.YamboPy-pol_reconstructed_F{i_f + 1}'
                header="[fs] Px Py Pz"
                if self.l_out_current:
                    output_file = f'o{self.prefix}.YamboPy-curr_reconstructed_F{i_f + 1}'
                    header="[fs] Jx Jy Jz"
                if (to_file):
                    np.savetxt(output_file, values, header=header, delimiter=' ', footer="Reconstructed signal")
                else:
                   return values




"""
This script computes the phonon-assisted luminescence spectra for indirect materials
"""
import numpy as np
import matplotlib.pyplot as plt
from yambopy.exciton_phonon.excph_luminescence import exc_ph_luminescence
from yambopy.exciton_phonon.excph_input_data import exc_ph_get_inputs

path = '3D_hBN'
# Path to BSE calculation (Lout--> response is Lfull)
bsepath =  f'{path}/bse_Lfull' # ndb.BS_diago_Q* databases are needed
# Path to BSE calculation for optically active exciton (Lin --> response is Lbar)
bseBARpath =  f'{path}/bse_Lbar'  # ndb.BS_diago_Q1 database is needed
# Path to electron-phonon calculation
elphpath = path # ndb.elph is needed
# Path to unprojected dipoles matrix elements (optional)
dipolespath = bsepath # ndb.dipoles is needed (optional)
# Path to lattice and k-space info
savepath = f'{path}/SAVE' # ns.db1 database is needed

bands_range=[6,10] # 2 valence, 2 conduction bands
phonons_range=[0,12] # All phonons
nexc_out = 12 # 12 excitonic states at each momentum (Lout)
nexc_in  = 12 # 12 excitonic states at Q=0 (Lin)
T_ph  = 10 # Lattice temperature
T_exc = 10 # Effective excitonic temperature

emin=4.4      # Energy range and plot details (in eV)
emax=4.7
estep=0.0002
broad = 0.005 # Broadening parameter for peak width (in eV)

# We calculate and load all the inputs:
# * Exciton-phonon matrix elements
# * Excitonic dipole matrix elements
# * Exciton energies
# * Phonon energies
# * We specify bse_path2=bseBARpath meaning we use Lbar calculation for Q=0 excitons
input_data = exc_ph_get_inputs(savepath,elphpath,bsepath,\
                               bse_path2=bseBARpath,dipoles_path=dipolespath,\
                               nexc_in=nexc_in,nexc_out=nexc_out,\
                               bands_range=bands_range,phonons_range=phonons_range)

ph_energies, exc_energies, exc_energies_in, G, exc_dipoles = input_data

# We calculate the luminescence spectrum including the input data from before 
w,PL = exc_ph_luminescence(T_ph,ph_energies,exc_energies,exc_dipoles,G,\
                           exc_energies_in=exc_energies_in,exc_temp=T_exc,\
                           nexc_out=nexc_out,nexc_in=nexc_in,emin=emin,emax=emax,\
                           estep=estep,broad=broad) #,ph_channels='e')

fig = plt.figure()
ax = fig.add_subplot(1,1,1)
ax.set_xlim(emin,emax)
ax.set_ylim(0,np.max(PL)*1.1)
ax.get_yaxis().set_visible(False)

ax.plot(w, PL, '-',c='red', label="AA' hBN luminescence")

plt.legend()
plt.savefig('hBN_luminescence.png')
plt.show()




import sys
import numpy as np
import scipy as special
import math  
from yambopy.units import speed_of_light,speed_of_light_SI,WMm22kWCMm2,WMm22ERGCMm2SECm1,AU2KWCMm2,FREE_SPACE_PERM,SEC2AU,fs2aut

def get_Efield_w(freqs,efield):
    if efield["name"] == "DELTA":
        efield_w=efield["amplitude"]*np.exp(1j*freqs[:]*efield["initial_time"])
    else:
        print("Fields different from Delta function not implemented yet")
        sys.exit(0)

    return efield_w

    
def Divide_by_the_Field(efield,order):
    
    if efield['name']=='SIN' or efield['name']=='SOFTSIN':
        if order !=0:
            divide_by_field=np.power(-2.0*1.0j/efield['amplitude'],order,dtype=np.cdouble)
        elif order==0:
            divide_by_field=-4.0/np.power(efield['amplitude'],2.0,dtype=np.cdouble)
    elif efield['name'] == 'QSSIN':
        # Approximate relations/does not work yet
        sigma=efield['width']
        W_0=efield['freq_range'][0]
        T_0= np.pi/W_0*float(round(W_0/np.pi*3.*sigma))
        T = 2*np.pi/W_0
        E_w= math.sqrt(np.pi/2)*sigma*np.exp(-1j*W_0*T_0)*(special.erf((T-T_0)/math.sqrt(2.0)/sigma)+special.erf(T_0/math.sqrt(2.0)/sigma))
    else:
        raise ValueError("Electric field not implemented in Divide_by_the_Field!")

    return divide_by_field

def Gaussian_centre(efield):
    ratio=np.pi/efield['freq_range'][0]
    sigma=efield["damping"]/(2.0*(2.0*np.log(2.0))**0.5)
    peak_ = float(1.0 /ratio * sigma *efield["peak"])
    return ratio * float(round(peak_)),sigma

def Efield_strength(Intensity, unit_system):
    """
    From Octopus ( http://www.tddft.org/programs/octopus )

    It is very common to describe the strength of a laser field by its intensity, 
    rather than using the electric field amplitude. In atomic units, the relationship 
    between instantaneous electric field and intensity is:

        I(t) = c / (8π) * E^2(t)

    It is common to read intensities in kW cm^-2. The dimensions of intensities are [W]/(L^2T), 
    where [W] are the dimensions of energy. The relevant conversion factors are:

        Hartree / (a_0^2 * atomic_time) = 6.4364086e+15 W / cm^2 = 6.4364086e+12 kW / cm^2

    In Yambo AU2KWCMm2 = 6.4364086e+12

    --------------------------------------------------------------------------------------------
    This simple function uses the formula valid in SI and CGS to extract 
    the field intensity given in atomic units.
    """

    # Work space
    SPEED = 0.0
    I = 0.0

    # From Boyd, "Nonlinear Optics", 3rd edition, page 602–603
    # Assuming n=1 (epsilon=1)

    if unit_system == "SI":
        # Convert input intensity to correct SI-based scale
        I = Intensity * AU2KWCMm2 / WMm22KWCMm2

        SPEED = speed_of_light_SI

        # I = 1 / (FREE_SPACE_PERM * SPEED_OF_LIGHT) * |E|^2 
        Efield_strength = math.sqrt(I / FREE_SPACE_PERM / SPEED) * VMm12AU

    elif unit_system == "CGS":
        # Convert input intensity to correct CGS-based scale
        I = Intensity * AU2KWCMm2 / WMm22KWCMm2 / WMm22ERGCMm2SECm1

        SPEED = speed_of_light_SI * 100.0  # cm/sec

        # I = SPEED_OF_LIGHT / (4π) * |E|^2 
        Efield_strength = math.sqrt(I * 4.0 * math.pi / SPEED) * SVCMm12VMm1 * VMm12AU

    elif unit_system == "AU":
        # In atomic units, use direct formula
        I = Intensity
        SPEED = speed_of_light

        # I = SPEED_OF_LIGHT / (4π) * |E|^2 
        Efield_strength = math.sqrt(I * 4.0 * math.pi / SPEED)

    else:
        raise ValueError(f"Unknown unit system: {unit_system}")

    return Efield_strength

def theta_function(T, step, order):
    """
    Element-wise version of the Fortran theta_function.
    
    Parameters
    ----------
    T : float or np.ndarray
        Input variable (can be scalar or array).
    step : float
        Step size.
    order : int
        Order of the derivative (0, 1, or 2).
        
    Returns
    -------
    float or np.ndarray
        Value of the theta function (or its derivative) evaluated at T.
    """
    T = np.asarray(T, dtype=float)
    theta = np.zeros_like(T)

    # Case where |T/step| > 1.05
    mask_far = np.abs(T / step) > 1.05
    if order == 0:
        theta[mask_far & (T > 0)] = 1.0
    # Higher orders remain zero by default
    if np.all(mask_far):
        return theta

    # Determine step_case
    step_cases = np.full(T.shape, "", dtype="<U4")
    x = T / step
    step_cases[np.abs(x + 1.0) < 0.1] = "-1.0"
    step_cases[np.abs(x + 0.5) < 0.1] = "-0.5"
    step_cases[np.abs(x + 0.0) < 0.1] = " 0.0"
    step_cases[np.abs(x - 0.5) < 0.1] = "+0.5"
    step_cases[np.abs(x - 1.0) < 0.1] = "+1.0"

    # Apply the cases
    for case in ["-1.0", "-0.5", " 0.0", "+0.5", "+1.0"]:
        mask = step_cases == case
        if not np.any(mask):
            continue
        if case == "-1.0":
            if order == 2:
                theta[mask] = 1.0 / step**2
        elif case == "-0.5":
            if order == 0:
                theta[mask] = 0.25
            elif order == 1:
                theta[mask] = 0.50 / step
            elif order == 2:
                theta[mask] = 1.00 / step**2
        elif case == " 0.0":
            if order == 0:
                theta[mask] = 0.5
            elif order == 1:
                theta[mask] = 1.0 / step
            elif order == 2:
                theta[mask] = 0.0
        elif case == "+0.5":
            if order == 0:
                theta[mask] = 0.75
            elif order == 1:
                theta[mask] = 0.50 / step
            elif order == 2:
                theta[mask] = -1.00 / step**2
        elif case == "+1.0":
            if order == 0:
                theta[mask] = 1.0
            elif order == 2:
                theta[mask] = -1.00 / step**2

    return theta



#
# License-Identifier: GPL
#
# Copyright (C) 2024 The Yambo Team
#
# Authors: HPC, FP
#
# This file is part of the yambopy project
#
import numpy as np
from itertools import product

def calculate_distances(kpoints):
    """
    take a list of k-points and calculate the distances between all of them
    """
    kpoints = np.array(kpoints)
    distances = [0]
    distance = 0
    for nk in range(1,len(kpoints)):
        distance += np.linalg.norm(kpoints[nk-1]-kpoints[nk])
        distances.append(distance)   
    return distances
 
def expand_kpts(kpts,syms):
    """ 
    Fast expansion giving only coordinate list in output. See kpoints.py for a more complete version.

    Take a list of qpoints and symmetry operations and return the full brillouin zone
    with the corresponding index in the irreducible brillouin zone
    """
    full_kpts = []
    print("nkpoints:", len(kpts))
    for nk,k in enumerate(kpts):
        for sym in syms:
            full_kpts.append((nk,np.dot(sym,k)))

    return full_kpts

def vec_in_list(veca,vec_list,atol=1e-6):
    """
    Check if a vector exists in a list of vectors
    """
    return np.array([ np.allclose(veca,vecb,rtol=atol,atol=atol) for vecb in vec_list ]).any()


def isbetween(a,b,c,eps=1e-5):
    """ Check if c is between a and b
    """
    return np.isclose(np.linalg.norm(a-c)+np.linalg.norm(b-c)-np.linalg.norm(a-b),0,atol=eps)

def red_car(red,lat):
    """
    Convert reduced coordinates to cartesian
    """
    return np.array([coord[0]*lat[0]+coord[1]*lat[1]+coord[2]*lat[2] for coord in red])

def car_red(car,lat):
    """
    Convert cartesian coordinates to reduced
    """
    return np.array([np.linalg.solve(np.array(lat).T,coord) for coord in car])

def vol_lat(lat):
    """
    Calculate the volume of a lattice
    """
    a1,a2,a3 = np.array(lat)
    return np.dot(a1,np.cross(a2,a3))

def rec_lat(lat):
    """
    Calculate the reciprocal lattice vectors
    """
    v = vol_lat(lat)
    a1,a2,a3 = np.array(lat)
    b1 = np.cross(a2,a3)/v
    b2 = np.cross(a3,a1)/v
    b3 = np.cross(a1,a2)/v
    return np.array([b1,b2,b3])

def replicate_red_kmesh(kmesh,repx=list(range(1)),repy=list(range(1)),repz=list(range(1))):
    """
    copy a kmesh in the tree directions
    the kmesh has to be in reduced coordinates
    """
    kmesh = np.array(kmesh)
    kmesh_nkpoints = len(kmesh)

    kmesh_full = []
    kmesh_idx  = []
    for x,y,z in product(repx,repy,repz):
        kmesh_shift = kmesh + np.array([x,y,z])
        kmesh_full.append(kmesh_shift)
        kmesh_idx.append(list(range(kmesh_nkpoints)))

    return np.vstack(kmesh_full), np.hstack(kmesh_idx)


def point_matching(a,b,double_check=True,debug=False,eps=1e-8):
    """
    Matches the points of list a to the points of list b
    using a nearest neighbour finding algorithm

    Arguments:

        double_check: after the nearest neighbours are assigned check further
        if the distance between points is within the precision eps

        eps: precision for the double check (default: 1e-8)

    """
    #karma
    from scipy.spatial import cKDTree
    from time import time
    a = np.array(a)
    b = np.array(b)
    start_time = time()

    #initialize the kdtree
    kdtree = cKDTree(a, leafsize=10)
    map_b_to_a = []
    for xb in b:
        current_dist,index = kdtree.query(xb, k=1, distance_upper_bound=6)
        map_b_to_a.append(index)
    map_b_to_a = np.array(map_b_to_a)
    
    if debug:
        print("took %4.2lfs"%(time()-start_time))

    if double_check:
        for ib,ia in enumerate(map_b_to_a):
            dist = np.linalg.norm(a[ia]-b[ib])
            if dist > eps:
                raise ValueError('point a %d: %s is far away from points b %d: %s  dist: %lf'%(ia,str(a[ia]),ib,str(b[ib]),dist))

    return map_b_to_a

def bravais_types(lats,alat_0):
    """
    Determine Bravais lattice type of unit cell

    :: lats -> lattice vectors from YamboLatticeDB corresponding to lat
    :: alat_0 -> lattice parameter from YamboLatticeDB corresponding to alat[0]

    More lattice types to be implemented
    """
    from math import sqrt,cos,sin

    bravais_types = ['Hexagonal and Trigonal P','Orthorhombic P']

    lats_ = lats/alat_0
 
    if np.array_equal(lats_[0],[1.,0.,0.]):
        
        if np.allclose(lats_[1],[-0.5,sqrt(3.)/2.,0.]):
        
            if np.allclose(lats_[2],[0.,0.,lats_[2,2] ]): return bravais_types[0]

        if np.array_equal(lats_[1],[0.,lats_[1,1],0.]):
         
            if np.allclose(lats_[2],[0.,0.,lats_[2,2] ]): return bravais_types[1]
    else: return 'No type'



from qepy import Path,ProjwfcXML,ProjwfcIn
from math import sqrt

# Matplotlib options
import matplotlib.pyplot as plt
fig = plt.figure(figsize=(5,7))
ax  = fig.add_axes( [ 0.12, 0.10, 0.70, 0.80 ])

# k-points map
npoints = 50
path_kpoints = Path([ [[0.0, 0.0, 0.0],'$\Gamma$'],
                      [[0.5, 0.0, 0.0],'M'],
                      [[1./3,1./3,0.0],'K'],
                      [[0.0, 0.0, 0.0],'$\Gamma$']], [int(npoints*2),int(npoints),int(sqrt(5)*npoints)])

# Class Projwfc
# Class to run projwfc.x and create 
# atomic_proj.xml (comment if already done)
'''
proj = ProjwfcIn(prefix='bn')
proj.run(folder='bands')
'''

# Class ProjwfcXML
# Atom-projected band structure
band = ProjwfcXML(prefix='bn',path='bands')
# print info
print(band)

# Manual selection of the lists of states by inspecting projwfc output
#atom_1 = list(range(16))
#atom_2 = list(range(16,32))

# Automatic selection of the states
atom_1 = band.get_states_helper(atom_query=['N'])
atom_2 = band.get_states_helper(atom_query=['B'])

# Add scissor operator to the bands from a G0W0 calculation
#scissor= [1.8985195950522469, 1.0265240811345133, 1.051588659878575]
#n_val = 4
#band.add_scissor(n_val,scissor)

band.plot_eigen(ax,path_kpoints=path_kpoints,cmap='viridis',selected_orbitals=atom_1,selected_orbitals_2=atom_2)

# Plot colormap
#
import matplotlib as mpl
cmap=plt.get_cmap('viridis')
bx  = fig.add_axes( [ 0.88, 0.10, 0.05, 0.80 ])
norm = mpl.colors.Normalize(vmin=0.,vmax=1.)
cb1 = mpl.colorbar.ColorbarBase(bx, cmap=cmap, norm=norm,orientation='vertical',ticks=[0,1])
cb1.set_ticklabels(['B', 'N'])

plt.show()



# 
# License-Identifier: GPL
#
# Copyright (C) 2024 The Yambo Team
#
# Authors: JC-V
# First verion by JC-V (2024)
#
# This file is part of the yambopy project
#
#

from qepy import *
from yambopy import *
from re import findall
from scipy.interpolate import Rbf

class Spin_texture():

    def __init__(self,prefix,path='.',folder_spin='.',prefix_spin='Spin'):

        """ 
        Reading of spin-projected values and calculation of the spin texture as a colormap, where the spin-z is represented by the intensity

        === Usage and variables ===

        === Reading spin-projected files ===
        >> Spin_data = Spin_texture(prefix=db_prefix,path=db_path,folder_spin=db_folder_spin,prefix_spin=db_prefix_spin)
        >> Spin_{index} = Spin_data.load_spin_data(folder_spin=db_folder_spin, prefix_spin=db_prefix_spin, index)
 
        === Plot texture ===
        >> Spin_data = Spin_texture(prefix=db_prefix,path=db_path,folder_spin=db_folder_spin,prefix_spin=db_prefix_spin)
        >> Spin_data.plot_spin_texture(ax,limfactor=0.8,mode="raw",nband=db_nband)

        Input:
        :: prefix is the prefix of the QE directory
        :: path is the path of the QE save folder
        :: folder_spin is the path of the QE spin folder
        :: prefix_spin is the prefix of the spin files
        :: index represents 1,2,3 = x,y,z
        
        Output:
        :: Plot of the spin texture in three different ways: 
           - "raw": The data is represented as points directly from QE 
           - "interpolated": The data is interpolated from QE and subsequently represented
           -  "arrow": The data is represented from QE as arrows, where the direction of the arrows is defined by the spin-x and spin-y components
        """

        print("=== Initializing the data ===")

        # Prefix and path of QE data
        self.prefix = prefix
        self.path = path
        self.folder_spin = folder_spin
        self.prefix_spin = prefix_spin

        # Loading data from QE
        self.data_xml = PwXML(prefix=self.prefix, path=self.path)
        self.nkpoints = self.data_xml.nkpoints
        self.nbands = self.data_xml.nbands
  
    def load_spin_data(self, index):

        print(f"=== Reading spin-{index} data from files ===")

        # Reading spin files
        filepath = f'{self.folder_spin}/{self.prefix_spin}.out.{index}'
        with open(filepath, 'r') as file:
            lines = file.readlines()

        # Defining the structure to read the data properly
        nband = int(findall(r"[-+]?\d*\.\d+|\d+", lines[0].strip().split()[2])[0])
        nk = int(lines[0].strip().split()[-2])
        nline = nband // 10

        # Warnings
        if nband < 10 or nband % 10 != 0:
            raise ValueError("Error: nband must be >= 10 and a multiple of 10.")
        if self.nbands != nband or self.nkpoints != nk:
            print("Error: Dimensions are inconsistent!")

        # Definition of spin array
        spin = np.zeros((self.nkpoints, self.nbands))

        # Reading spin values and saving them in IBZ repetition
        for ik in range(self.nkpoints):
            for ib in range(nline):
                ib1, ib2 = int(ib * 10), int((ib + 1) * 10)
                line_idx = int(ik * (nline + 1) + 2 + ib)
                spin[ik, ib1:ib2] = list(map(float, lines[line_idx].split()))

        return spin 

    def plot_spin_texture(self, ax, limfactor=0.8, mode="raw", nband=1, **kwargs):

        print("=== Plotting spin texture ===")

        # Repetition of IBZ
        kpoints_cart = np.array(self.data_xml.kpoints)
        kpoints_red = np.vstack([car_red([k], self.data_xml.rcell)[0] for k in kpoints_cart])
        kmesh_full, kmesh_idx = replicate_red_kmesh(kpoints_red, repx=range(-1, 2), repy=range(-1, 2))
        self.x, self.y = red_car(kmesh_full, self.data_xml.rcell)[:, :2].T

        # Defining plot boundaries
        lim = np.max(self.data_xml.rcell) * limfactor

        # Loading spin-z data
        self.spin_projection = {}
        self.spin_projection[3] = self.load_spin_data(3)

        # Redefining spin-z in IBZ repetition
        self.spin_z = self.spin_projection[3][kmesh_idx]

        # Types of plot
        if mode == "raw":

            ax.scatter(self.x, self.y, c=self.spin_z[:, nband], cmap='seismic', s=20)

        elif mode == "interpolated":

            npts = kwargs.get('npts', 100)
            rbfi = Rbf(self.x, self.y, self.spin_z[:, nband], function='linear')
            xi = yi = np.linspace(-lim, lim, npts)
            zi = np.array([[rbfi(x, y) for x in xi] for y in yi])
            ax.imshow(zi, interpolation=kwargs.get('interp_method', 'bicubic'),
                      extent=[-lim, lim, -lim, lim], cmap='seismic')

        elif mode == "arrow":

            # Loading spin-x and spin-y data
            for index in range(1, 3):
                self.spin_projection[index] = self.load_spin_data(index)

            # Redefining spin-x and spin.y in IBZ repetition
            self.spin_x, self.spin_y = self.spin_projection[1][kmesh_idx], self.spin_projection[2][kmesh_idx]

            ax.quiver(self.x, self.y, self.spin_x[:, nband], self.spin_y[:, nband],
                      self.spin_z[:, nband], cmap='seismic', scale=20)

        else:
            raise ValueError("Invalid mode. Choose 'raw', 'interpolated', or 'arrow'.")

        # Defining plot boundaries
        ax.set_title(f"Spin texture - band = {nband}")
        ax.set_xlim(-lim, lim)
        ax.set_ylim(-lim, lim)
        ax.set_xlabel("k$_{x}$ (bohr$^{-1}$)")
        ax.set_ylabel("k$_{y}$ (bohr$^{-1}$)")

        print("=== Spin texture plotted successfully ===")



from netCDF4 import *
import argparse
import numpy as np

"""
Script to convert RL number in Ry energy units using ndb.gops.

Inputs:
 1. --ndb_gops path/to/SAVE/folder [or any folder containing ndb.gops]
 2. --value value to convert with units [e.g., 11 RL or 5 Ry]

The script will read ndb.gops and find the nearest completed G-shell, then give the
converted value in Ry (RL) to the one supplied in input.

"""

def convert(value,ndb_gops):
    """ Perform the conversion
    """ 

    def get_closed_shell(value,shells):
        """ Returns index of closest completed shell to input value
        """
        ind = np.searchsorted(shells, value, side='left')
        if ind==len(shells): return None
        else:                return ind

    Ry2Ha=0.5
    ndb_gops = ndb_gops+"/ndb.gops"
    try:    value, unit = value
    except: raise ValueError("[ERROR] Incorrect format of value to be converted")

    # Read shells from ndb.gops
    db = Dataset(ndb_gops,'r')
    ng_in_shell = np.array(db['ng_in_shell'][:]).astype(int)
    E_of_shell  = db['E_of_shell'][:]
    db.close()

    if unit=='RL': 
        value=int(value)
        shells_i = ng_in_shell
        shells_o = E_of_shell
        unit_o   = 'Ry'
    elif unit=='Ry': 
        value=float(value)*Ry2Ha
        shells_i = E_of_shell
        shells_o = ng_in_shell
        unit_o   = 'RL'
    else: raise ValueError("[ERROR] Unit %s currently not supported"%unit)

    ind = get_closed_shell(value,shells_i)
    if ind is not None:
        close_shell_value = shells_i[ind]
        converted_value   = shells_o[ind]
        print(converted_value,' ',unit_o,' (closest closed shell)')
    else: raise ValueError("[ERROR] value supplied is bigger than maximum shell at %s %s"%(str(shell_i[-1]),unit))

if __name__ == "__main__":
    parser = argparse.ArgumentParser(description='Convert RL number in energy units (Ry)')
    parser.add_argument('-gops','--ndb_gops', type=str, default="./SAVE", help='<Optional> Path to folder with ndb.gops (default: ./SAVE)')
    parser.add_argument('-v','--value', type=str,help="<Required> Value to be converted along with units, e.g.: '11 RL' or '5 Ry'",nargs=2,required=True)
    args = parser.parse_args()

    ndb_gops = args.ndb_gops
    value    = args.value
    
    convert(value,ndb_gops)



#
# Author: Henrique Pereira Coutada Miranda
# Run a IP calculation using yambo
#
from __future__ import print_function
import sys
from yambopy import *
from qepy import *
import argparse
from schedulerpy import *
import matplotlib.pyplot as plt

#parse options
parser = argparse.ArgumentParser(description='Test the yambopy script.')
parser.add_argument('-dg','--doublegrid', action="store_true", help='Use double grid')
parser.add_argument('-c', '--calc', action="store_true", help='calculate the IP absorption')
parser.add_argument('-p', '--plot', action="store_true", help='plot the results')
args = parser.parse_args()

if len(sys.argv)==1:
    parser.print_help()
    sys.exit(1)

yambo  = 'yambo'
p2y    = 'p2y'
ypp    = 'ypp'
folder = 'ip'
prefix = 'bn'
scheduler = Scheduler.factory

#check if the SAVE folder is present
if not os.path.isdir('database/SAVE'):
    print('preparing yambo database')
    p2y_run = scheduler()
    p2y_run.add_command('mkdir -p database')
    p2y_run.add_command('cd nscf/bn.save; %s > %s.log'%(p2y,p2y))
    p2y_run.add_command('%s > %s.log'%(yambo,yambo))
    p2y_run.add_command('mv SAVE ../../database/')
    p2y_run.run()

if not os.path.islink('%s/SAVE'%folder):
    s = scheduler()
    s.add_command('mkdir -p %s'%folder)
    s.add_command('cd %s; ln -s ../database/SAVE .'%folder)
    if not args.doublegrid: s.add_command('cd .. ; rm -f database/SAVE/ndb.Double_Grid')
    s.run()

#initialize the double grid
if args.doublegrid and not os.path.isfile('database/SAVE/ndb.Double_Grid'):

    #check if the double grid nscf cycle is present
    if os.path.isdir('nscf_double/%s.save'%prefix):
        print('nscf_double calculation found!')
    else:
        print('nscf_double calculation not found!')
        exit()

    if not os.path.isdir('database_double/SAVE'):
        print('preparing yambo double database')
        shell = scheduler()
        shell.add_command('cd nscf_double/%s.save; %s; %s'%(prefix,p2y,yambo))
        shell.add_command('cd ../../')
        shell.add_command('mkdir -p database_double')
        shell.add_command('mv nscf_double/%s.save/SAVE database_double'%prefix)
        shell.run()

    #initialize the double grid
    print("creating double grid")

    yppin = YamboIn.from_runlevel('%s -m',filename='ypp.in',executable=ypp,folder='database')

    yppin['DbGd_DB1_paths'] = ["../database_double"]
    yppin.arguments.append('SkipCheck')

    yppin.write('database/ypp.in')

    shell = scheduler()
    shell.add_command('cd database; %s'%ypp)
    shell.add_command('cd ../%s ; rm -rf yambo o-*'%folder)
    #print(shell)
    shell.run()

if args.calc:
    #create the yambo input file
    y = YamboIn.from_runlevel('%s -o g -V all'%yambo,folder=folder)

    y['FFTGvecs'] = [30,'Ry']
    y['BndsRnXs'] = [1,30]
    y['QpntsRXd'] = [[1,1],'']
    y['ETStpsXd'] = 500

    y.write('%s/yambo_run.in'%folder)

    print('running yambo')
    yambo_run = scheduler()
    yambo_run.add_command('cd %s; %s -F yambo_run.in -J yambo'%(folder,yambo))
    if args.doublegrid: yambo_run.add_command('cd ..; rm database/SAVE/ndb.Double_Grid')
    yambo_run.run()

if args.plot:
    # Plot absorption spectrum
    data=np.genfromtxt('%s/o-yambo.eps_q1_ip'%folder,usecols=(0,1))
    fig = plt.figure(figsize=(4,5))
    ax = fig.add_axes( [ 0.20, 0.20, 0.70, 0.70 ])

    plt.plot(data[:,0],data[:,1],'-',c='b',label='IP Absorption')
    plt.legend()

    plt.show()



#
# Author: MN
#
import os
import numpy as np
from yambopy.dbs.excitondb import YamboExcitonDB
from yambopy.dbs.latticedb import YamboLatticeDB
from yambopy.dbs.wfdb import YamboWFDB
from .exciton_matrix_elements import exciton_X_matelem
from yambopy.tools.degeneracy_finder import find_degeneracy_evs
from yambopy.tools.citations import citation

@citation("A. R. Kshirsagar et al. Phys. Rev. B 112, 12 (2025)")
def compute_exciton_spin(lattice, excdb, wfdb, elec_sz, contribution='b',diagonal=False):
    """
    Compute the spin matrix elements <S'|S_z|S> for excitons.

    This function calculates the spin matrix elements for excitons using the
    wavefunctions and spin operators. The spin matrix is computed in the basis
    of exciton states, and off-diagonal elements are included. Diagonalization
    of the matrix in degenerate subspaces is required to obtain the spin values.

    Parameters
    ----------
    lattice : latticedb
        Lattice database
    excdb : exciton db
        Exciton Database 
    wfdb : wfc db
        wavefunction Database 
    elec_sz : ndarray
        Electron spin matrix elements (nk, nbnds. nbnds).
    contribution : str, optional
        Specifies which contribution to compute:
        - 'b': Total spin (default).
        - 'e': Electron spin only.
        - 'h': Hole spin only.
    diagonal : bool, optional
        If True, only diagonal spin elements are computed. Default is False.

    Returns
    -------
    exe_Sz : ndarray
        Spin matrix elements for excitons with shape (nstates, nstates).
    """
    #
    # Ensure the calculation is valid only for spinor wavefunctions
    assert wfdb.nspinor == 2, "Makes sense only for nspinor = 2"
    #
    # Sanity check
    assert np.min(excdb.table[:, 1]) - 1 == wfdb.min_bnd, \
            "wfdb and exciton db are inconsistant (Bands)"
    ## sanity check
    assert np.max(excdb.table[:, 2]) == wfdb.min_bnd + wfdb.nbands, \
            "wfdb and exciton db are inconsistant (Bands)"
    #
    assert elec_sz.shape == (wfdb.nkBZ, wfdb.nbands, wfdb.nbands)
    # get Akcv
    Akcv = excdb.get_Akcv()
    #
    # Get the exciton q-point in Cartesian coordinates
    excQpt = excdb.car_qpoint
    #
    # Convert the q-point to crystal coordinates
    excQpt = lattice.lat @ excQpt
    #
    # Compute the exciton spin matrix elements <S'|S_z|S>
    exe_Sz = exciton_X_matelem(excQpt, np.array([0, 0, 0]), Akcv,
                               Akcv, elec_sz[None,:,None,...], wfdb.kBZ,
                               diagonal_only=diagonal,contribution=contribution)
    #
    return exe_Sz[0]





def compute_exc_spin_iqpt(path='.', bse_dir='SAVE', iqpt=1,
                          nstates=-1, contribution='b', degen_tol = 1e-2,
                          sz=0.5 * np.array([[1, 0], [0, -1]]),
                          return_dbs_and_spin=True):
    """
    
    
    Description
    -----------
    Compute expectation value of S_z operator for excitons.

    Parameters
    ----------
    path : str, optional
        Path to the directory containing calculation SAVE and BSE folder.
        Default: '.' (current directory)
    bse_dir : str, optional
        Directory containing BSE calculation data. Default: 'SAVE'
    iqpt : int or array-like, optional
        Q-point index or list of Q-point indices to analyze. Default: 1
        (Fortran indexing)
    nstates : int, optional
        Number of excitonic states to consider. Use -1 for all states. Default: -1
    contribution : str, optional
        Which contribution to compute:
        - 'b': both electron and hole (default)
        - 'e': electron only
        - 'h': hole only
    degen_tol : float, optional
        Tolerance for detecting degenerate states. Default: 1e-2
    sz : ndarray, optional
        S_z operator matrix representation. Default: 0.5 * np.array([[1, 0], [0, -1]])
    return_dbs_and_spin : bool, optional
        If True, returns both spin values and database objects. Default: True

    Returns
    -------
    exe_Sz : ndarray
        Array containing S_z expectation values for excitonic states
    dbs_objects : list, optional
        If return_dbs_and_spin=True, returns [lattice, wfdb, excdb, elec_sz] database objects
    Examples
    --------
    Compute the total spin matrix elements for excitons:

    >>> import numpy as np
    >>> from yambopy.bse.exciton_spin import compute_exc_spin_iqpt
    >>> Sz_exe = compute_exc_spin_iqpt(bse_dir='GW_BSE', nstates=2)
    >>> print(Sz_exe)

    Compute only the electron spin contribution:

    >>> Sz_exe = compute_exc_spin_iqpt(bse_dir='GW_BSE', nstates=2, contribution='e')

    Compute only the hole spin contribution:

    >>> Sz_exe = compute_exc_spin_iqpt(bse_dir='GW_BSE', nstates=2, contribution='h')
    """
    #
    ## Check if it single Q or multiple Q's
    if np.isscalar(iqpt): iqpt = [iqpt]
    else : iqpt = list(iqpt)
    # Load the lattice database
    lattice = YamboLatticeDB.from_db_file(os.path.join(path, 'SAVE', 'ns.db1'))
    ## load exbds
    excdb = []
    for iq in iqpt:
        filename = 'ndb.BS_diago_Q%d' % (iq)
        excdb.append(YamboExcitonDB.from_db_file(lattice, filename=filename,
                                                 folder=os.path.join(path, bse_dir),
                                                 Load_WF=True, neigs=nstates))
    # Load the wavefunction database
    wfdb = YamboWFDB(path=path, latdb=lattice,
                      bands_range=[np.min(excdb[0].table[:, 1]) - 1,
                    np.max(excdb[0].table[:, 2])])
    #
    # Compute the spin matrix elements in the BZ
    elec_sz = wfdb.get_spin_m_e_BZ(s_z=sz)
    #
    exe_Sz = []
    for ixdb in excdb:
        smat = compute_exciton_spin(lattice, ixdb,
                                           wfdb, elec_sz,
                                           contribution=contribution,
                                           diagonal=False)
        smat = get_spinvals(smat, ixdb.eigenvalues, atol=degen_tol)
        ss_tmp = []
        for i in smat: ss_tmp = ss_tmp + list(i)
        exe_Sz.append(ss_tmp)
    #
    exe_Sz = np.array(exe_Sz)
    if return_dbs_and_spin : return exe_Sz,[lattice, wfdb, excdb, elec_sz]
    else : return exe_Sz



def get_spinvals(spin_matrix, eigenvalues, atol=1e-3, rtol=1e-3):
    degen_idx = find_degeneracy_evs(eigenvalues,atol=atol, rtol=rtol)
    spins = []
    for id in degen_idx:
        w = np.linalg.eigvals(spin_matrix[id,:][:,id])
        spins.append(w)
    return spins






# Code from Myrta Gruning (2023)

import sys
import numpy as np
import matplotlib.pyplot as plt
from yambopy.units import fs2aut
from scipy import signal

def zeropadding_signal(f,a,Nval):
    step = f[1]-f[0]
    pf = np.zeros(2*Nval+len(f))
    pf[Nval-1:Nval+len(f)-1] = f[:]
    for n in range(Nval-1,-1,-1):
        pf[n] = pf[n+1] - step
    for n in range(Nval+1):
        pf[Nval+len(f)+n-1] = pf[Nval+len(f)+n-2] + step
#    pa = np.pad(a,(Nval,),'constant',constant_values=(a[0],a[-1])) #here I supposed that the first and last values are zero
    pa = np.pad(a,(Nval,),'constant',constant_values=(0,0)) #here I supposed that the first and last values are zero
    return pf,pa


def plot_signal(data=None,time=None,hdir=[1,0,0],padded=False,Npad=600,tstring='Signal',singlefig = True):
    #
    if data is None or time is None:
        print('Error you should provide an array with the data and time')
        sys.exit(0)
    
    hdir=hdir/np.linalg.norm(hdir)
    hdata=np.dot(data.T, hdir)

    if (singlefig):
        plt.show()
    plt.title(tstring)
    plt.xlabel('Time [fs]')
    if padded:
        nfreq,spadded=zeropadding_signal(time/fs2aut,hdata,Npad)
        plt.plot(nfreq,spadded)
    else:
        plt.plot(time/fs2aut,hdata)
    return

def get_psd(data=None, time=None,wind="blackman",hdir=[1,0,0],padded=False,Npad=600):
    #
    hdir=hdir/np.linalg.norm(hdir)
    hdata=np.dot(data.T, hdir)
    #
    if padded:
        nfreq,spadded=zeropadding_signal(time/fs2aut,hdata,Npad)
        f, Pxx_den = signal.periodogram(spadded,1.0/(nfreq[1]-nfreq[0]),scaling='spectrum',window=wind)
    else:
        f, Pxx_den = signal.periodogram(hdata,fs=fs2aut/(time[1]-time[0]),scaling='spectrum',window=wind)

# 1 hertz [Hz] = 4.13566553853599E-15 electron-volt [eV]

    return f*4.1356655385, Pxx_den/np.max(Pxx_den)

def plot_psd(f,psd,lfreq,tstring='',singlefig = True, lmax=41,ymin=10e-7, gap=None):
    l=np.arange(1,lmax)
    if (singlefig):
        plt.show()
        tlabel='Laser freq = '+str(lfreq)+'eV'
        ttitle = tstring
    else:
        tlabel = tstring
        ttitle = ''
    print(f)
    plt.semilogy(f/lfreq, psd,label=tlabel)
    plt.title(ttitle)
    plt.ylim([ymin,1])
    plt.xlim([0.1,lmax])
    plt.vlines(l,ymin,1,linestyles='dotted')
    plt.xticks(l)
    if gap is not None:
        plt.vlines(gap/lfreq,ymin,1,linestyles='dotted',colors='red')
    plt.xlabel('Harmonic number')
    plt.ylabel(r"PSD $\left[V^2\right]$")
    plt.legend()
    return




from yambopy.common.workflow import wait_for_job
import subprocess
import os
from schedulerpy import *
from copy import deepcopy

"""
This file contains the basic functions needed to check on calculations for 
completion and more.

TODO: Include a shell_run function for all executables

"""
def shell_qe_run(job_name,inp_name,out_name,run_dir,exec='pw.x',shell_name='qe',scheduler=None,depend_on_JOBID=None,hang_python=False,pre_run=[],pos_run=[]):
    """ 
    Submit QUANTUM ESPRESSO job
    
        exec: executable to be run with full path.
              options: /path/to/pw.x, /path/to/ph.x
        
        job_name: job name
        shell_name: name of *.sh script which is generated
        depend_on_JOBID: job id of simulation that the present job has a dependency on
        run_dir: where job is run
        out_name: name of output file
        inp_name: name of input file
        scheduler: instance of scheduler class (if not present, bash is initialised)
        pre_run / pos_run: LISTS containing commands to be added before / after the mpirun command
        hang_python: if True, python process sleeps until job is completed
        
        returns id of present submitted job (-1 if scheduler is bash)         
    """
    # Check executable
    if not exec[-4:]=='pw.x' and not exec[-4:]=='ph.x':
        raise ValueError('Executable not recognised (pw.x and ph.x are the only options).')
        
    # Copy scheduler instance in order to safely edit it
    if scheduler is None: shell = Scheduler.factory(scheduler="bash")
    else:                 shell = deepcopy(scheduler)
    
    shell.name = '%s_%s'%(job_name,shell.name)
    
    # Add dependency if specified
    if depend_on_JOBID is not None and shell.schedulertype != 'bash':
        dependency='afterok:%s'%depend_on_JOBID
        shell.kwargs['dependency']=dependency
        
    # Add additional commands if present
    if len(pre_run) != 0:
        for command in pre_run: shell.pre_run.append(command)

    # Additional commands to be executed after the main mpirun command
    if len(pos_run) != 0:   
        for command in pos_run: shell.pos_run.append(command)
 
    # Main mpirun command
    shell.add_mpirun_command('%s -inp %s > %s'%(exec,inp_name,out_name))
    shell.run(filename='%s/%s.sh'%(run_dir,shell_name)) ### Specify run path
    
    # Manage submissions if specified
    if hang_python: wait_for_job(shell,run_dir)
    if shell.schedulertype!='bash': this_job_id = shell.jobid
    else:                           this_job_id = -1
    
    shell.clean()
    
    return this_job_id

def check_qe_completed(folder,prefix,output_file,calc_type='pw'):
    """ 
    Check if qe calculation has correctly completed.
    
    - folder: where the calculation has been run.
    - prefix: qe prefix
    - output_file: name of output file
    - calc_type: either 'pw' or 'ph' or 'gkkp'
    
    """
    status = True
    
    # If save folder does not exist, return False (= NOT completed) immediately
    if calc_type=='pw' and not os.path.isdir('%s/%s.save'%(folder,prefix)):
        status = False
        return status
    elif calc_type=='ph' and not os.path.isdir('%s/_ph0'%folder):
        status = False
        return status
    elif calc_type=='gkkp' and not os.path.isdir('%s/elph_dir'%folder):
        status = False
        return status
    if calc_type != 'pw' and calc_type != 'ph' and calc_type != 'gkkp':
        raise ValueError("calc_type not recognised: it has to be either 'pw' or 'ph' or 'gkkp'.")
        
    # Next, check if output is correctly completed
    try:
        check = subprocess.check_output("grep JOB %s/%s*"%(folder,output_file), shell=True, stderr=subprocess.STDOUT)
        check = check.decode('utf-8')
        check = check.strip().split()[-1]
    except subprocess.CalledProcessError as e:
        check = ""
    if check != "DONE.": status = False
    return status



"""
Tutorial for YamboElectronPhononDB.

Electron-phonon matrix element reading and plotting

EDIT the path below to point to the yambo SAVE folder.
"""
save_path='BSE_saves/YAMBO_saves'
dipoles_path='BSE_saves/BSE_databases'
from yambopy import YamboLatticeDB,YamboDipolesDB,YamboElectronsDB
import numpy as np
import matplotlib.pyplot as plt

if __name__ == "__main__":
    
    Kspace_Plot     = True
    Absorption_Plot = False

    # Customly chosen  matrix elements
    i_c, i_v = [4,3] # 3-> top valence, 4-> bottom conduction
    i_x, i_y = [0,1] # Cartesian direction: 0 -> x, 1 -> y, 2 -> z

    #                    #
    # Start Yambopy part #
    #                    #

    # Create "lattice" object by reading the ns.db1 database inside the yambo SAVE
    ylat = YamboLatticeDB.from_db_file(filename=save_path+'/SAVE/ns.db1')

    # Read dipole matrix elements
    ydip = YamboDipolesDB.from_db_file(ylat,filename=f"{dipoles_path}/ndb.dipoles")
    print(ydip)

    # Read electron energies
    yel = YamboElectronsDB.from_db_file(folder=save_path+'/SAVE')
    print(yel)
    
    # Plot dipoles in k-space (modulus, summed over x,y,z)
    dip_of_k = np.sqrt( np.sum( np.abs(ydip.dipoles[:,:,i_c,i_v])**2., axis=1) )
    #dip_of_k = np.sqrt( np.abs(ydip.dipoles[:,i_x,i_c,i_v])**2.+np.abs(ydip.dipoles[:,i_y,i_c,i_v])**2. ) # sum over x,y plane only
    ydip.plot_dipoles(dip_of_k,s=100,plt_cbar=False,marker='H',cmap='viridis')

    # Get independent-particles absorption spectrum
    w, eps2 = ydip.ip_eps2(yel,ntot_dip=10,broad=0.01,emin=0.,emax=7.,esteps=1000)

    #                   #
    # End Yambopy part. #
    #                   #
    
    if Kspace_Plot:
        # Plot is customisable as needed using matplotlib
        ydip.ax.set_title('|d(k)| (a.u.)')
        ydip.ax.set_xlabel('k_x')
        ydip.ax.set_ylabel('k_y')
        plt.show()

    if Absorption_Plot:
        # Plot is customisable as needed using matplotlib
        plt.title('IP absorption spectrum')
        plt.xlim(0,7)
        plt.xlabel('Energy (eV)')
        plt.ylim(0,np.max(eps2)*1.1)
        plt.plot(w,eps2,color='blue')
        plt.show()




#########################################################
#  YAM(BO)PY(THON) Library
#
#  Generation of Yambo input files using python
#
#  Authors: A Molina-Sanchez, HPC Miranda
#
#  January 2016
#########################################################
# Calculation of spectra in the Independant Particle approximation
# Author: Alejandro Molina-Sanchez
#########################################################
# Collection of real-time scripts:
# cohsex-rt.py | bse-rt.py | merging-rt.py | ip-rt.py
#########################################################

#########################################################
from __future__ import print_function
from builtins import range
from yambopy     import *
from schedulerpy import *

############## SETTINGS ##############

yambo_module = 'yambo/master-intel'
yambo_rt     = 'yambo_rt'

source       = 'QSSIN-1e+03-70.0fs-1.94eV-0K'
folder_rt    = 'rt-24x24'

ip_nodes =  1
ip_cores =  12

time_probe = list(range(0,610,150))
#time_probe=(0,)
print(time_probe)


ip = YamboIn('%s -o b'%yambo_rt,folder=folder_rt) # NEQ COHSEX
# Common variables
ip['DBsIOoff'] = 'DIP'
ip['FFTGvecs'] = [ 20   , 'Ha'  ]
ip['BSKmod']   = 'IP'
ip['BSENGexx'] = [ 20   , 'Ha'  ]
ip['BSENGBlk'] = [ 1000 , 'mHa' ]
ip['NGsBlkXs'] = [ 1000 , 'mHa' ]
ip['BndsRnXs'] = [1 ,60]
ip['BSEBands'] = [25,28]
#paralelization
ip['X_all_q_ROLEs'] = 'q.k.c.v'
ip['X_all_q_CPU']   = '1.12.1.1'
ip['BS_ROLEs']      = 'k.eh.t'
ip['BS_CPU']        = '12.1.1'
ip['BSKmod']        = "IP"
ip['BEnRange'] =  [ [0.0  , 4.0] , 'eV'  ]    # Energy range spectra
ip['BLongDir'] =  [ 1.0, 0.0, 0.0 ]           # [BSS] [cc] Electric Field
ip['BDmRange'] =  [ [0.100 , 0.100] , 'eV'  ]   # Width
ip['BEnSteps'] =    1000                      # Energy steps
#ip.arguments.append('EvalKerr')
#ip.arguments.append('ForceEqTrans')


# Submission of the jobs for each time

print('Running IP calculations...')
print('RT source calculation: %s \n' % source)

for time in time_probe:
    print('Time of carriers database %d' % time)
    ip['XfnRTdb'] = 'f @ %d fs < ./pulse/ndb.RT_carriers' % ( time )
    ip['KfnRTdb'] = 'f @ %d fs < ./pulse/ndb.RT_carriers' % ( time )
    ip['RfnRTdb'] = 'f @ %d fs < ./pulse/ndb.RT_carriers' % ( time )
    nameip        = 'IP-t%d'              % ( time )
    print(nameip)
    ip.write('%s/%s/%s.in' %(folder_rt, source, nameip))
    yambo = oarsub(core=ip_cores,dependent=0,name='ip-rt',walltime="02:00:00")
    yambo.add_command('module load %s'%yambo_module)
    yambo.add_command('export OMP_NUM_THREADS=1')
    yambo.add_command('cd %s/%s ; mpirun -hostfile \$OAR_NODEFILE %s -F %s.in -J %s -C %s -I \'../\''%(folder_rt,source,yambo_rt,namebs,namebs,namebs))
    yambo.write('%s/%s/%s.ll' % (folder_rt, source, namebs))
    yambo.run()



#########################################################
#  YAM(BO)PY(THON) Library
#
#  Generation of Yambo input files using python
#
#  Authors: A Molina-Sanchez, HPC Miranda
#
#  January 2016
#########################################################
# Calculation of COHSEX corrections for different times
# in real-time calculations
# Author: Alejandro Molina-Sanchez
#########################################################
# Collection of real-time scripts:
# cohsex-rt.py | bse-rt.py | merging-rt.py | ip-rt.py
#########################################################

#########################################################
from __future__ import print_function
from builtins import range
from yambopy     import *
from schedulerpy import *


############## SETTINGS ##############

yambo_module = 'yambo/intel-4.1'
yambo_rt     = 'yambo_rt'

folder_rt    = 'rt-6x6'
source       = 'QSSIN-1e+03-70.0fs-2.0eV-0K'

CSRTmode = 'XG'  #  X: Screening, G: GFs

nodes =  1
cores =  12

time_probe = list(range(0,610,150))
#time_probe=(0,)
print(time_probe)


#######################################

cs = YamboIn('%s -r -p c -g n'%yambo_rt,folder=folder_rt) # NEQ COHSEX
#coulomb cutoff
cs['DBsIOoff'] = 'DIP'
cs['RandQpts'] =  1000000
cs['RandGvec'] =  [ 1, 'RL' ]
cs['CUTGeo']   =  "box z"
cs['CUTBox']   =  [ 0, 0, 38]
# Common variables
cs['FFTGvecs'] = [ 20   , 'Ha'  ]
cs['EXXRLvcs'] = [ 20   , 'Ha'  ]
cs['NGsBlkXs'] = [ 1000 , 'mHa' ]
cs['BndsRnXs'] = [1,60]
cs['QPkrange'] = [1,576,25,28]
#paralelization
cs['X_all_q_ROLEs'] = 'q.k.c.v'
cs['X_all_q_CPU']   = '1.%d.2.1'%(nodes*cores/2)
cs['SE_ROLEs']      = 'q.qp.b'
cs['SE_CPU']        = '1.%d.1'%(nodes*cores)
# Flags
cs.arguments.append('ExtendOut')


# Submission of the jobs for each time
print('Running COHSEX calculation...')
print('RT source calculation: %s \n' % source)

for time in time_probe:
    print('Time of carriers database %d' % time)
    if CSRTmode == 'X':
      cs['XfnRTdb'] = 'f @ %d fs < ./pulse/ndb.RT_carriers' % ( time )
    elif CSRTmode == 'G':
      cs['GfnRTdb'] = 'f @ %d fs < ./pulse/ndb.RT_carriers' % ( time )
    elif CSRTmode == 'XG':
      cs['XfnRTdb'] = 'f @ %d fs < ./pulse/ndb.RT_carriers' % ( time )
      cs['GfnRTdb'] = 'f @ %d fs < ./pulse/ndb.RT_carriers' % ( time )
    else:
      print('Error in the RT run level')
      exit()
    namecs        = 'C-%s-t%d'              % ( CSRTmode, time )
    print(namecs)
    cs.write('%s/%s/%s.in' %(folder_rt, source, namecs))
    yambo = oarsub(nodes=nodes,core=cores,dependent=0,name='cohsex',walltime="10:00:00")
    yambo.add_command('module load %s'%yambo_module)
    yambo.add_command('export OMP_NUM_THREADS=1')
    yambo.add_command('cd %s/%s ; mpirun -hostfile \$OAR_NODEFILE %s -F %s.in -J %s -C %s -I \'../\''%(folder_rt,source,yambo_rt,namecs,namecs,namecs))
    yambo.write('%s/%s/%s.ll' % (folder_rt, source, namecs))
    yambo.run()



# Copyright (C) 2018 Henrique Pereira Coutada Miranda, Alejandro Molina Sanchez
# All rights reserved.
#
# This file is part of yambopy
#
import os
import re
from math import sqrt

__all__ = ['PhIn']

class PhIn(object):
    """
    A class to generate an manipulate quantum espresso input files for ph.x
    """
    def __init__(self):
        self.variable = dict()
        self['tr2_ph']   = 1.0e-12
        self['prefix']   = '\"pw\"'
        self['fildyn']   = '\"dyn\"'
        self['qplot']    = '.false.'
        self['trans']    = '.true.'
        self.qpoints     = [] 

    @classmethod
    def from_qpoints(cls,qpoints):
        """
        Create a ph.x input from qpoints
        If qpoints is a list of lists write those qpoints explicit
        if qpoints is a list of indexes set the nq1,nq2 and nq3 indexes
        """
        instance = cls()
        if isinstance(qpoints[0],list): instance.qpoints = qpoints
        else: instance.set_nq(*qpoints)
        return instance

    @property
    def prefix(self):
        return self['prefix'].replace("'",'')

    @prefix.setter
    def prefix(self,value):
        self['prefix'] = "'%s'"%value.replace("'",'')

    @property
    def fildyn(self):
        return self['fildyn'].replace("'",'')

    @fildyn.setter
    def fildyn(self,value):
        self['fildyn'] = "'%s'"%value.replace("'",'')

    def write(self,filename):
        with open(filename,'w') as f:
            f.write(str(self))

    def set_vars(**kwargs):
        for var,value in kwargs.items():
            self[var] = value

    def set_nq(self,nq1,nq2,nq3):
        self['nq1'] = nq1
        self['nq2'] = nq2
        self['nq3'] = nq3
        self['ldisp'] = ".true."

    def __str__(self):
        lines = []; app = lines.append
        app("\n phonons \n&inputph")
        app(self.stringify_group('',self.variable)) #print variable
        if 'true' in self['qplot'].lower(): 
            app("%d"%len(self.qpoints))
            for q in self.qpoints:
                app("%12.8lf %12.8lf %12.8lf %d"%tuple(q))
        return "\n".join(lines)+"\n"

    def __setitem__(self,key,value):
        self.variable[key] = value

    def __getitem__(self,key):
        return self.variable[key]

    def stringify_group(self, keyword, group):
        if group != {}:
            lines = []; app = lines.append
            for keyword in group:
                app("%10s = %s" % (keyword, group[keyword]))
            app("/")
            return "\n".join(lines)
        else:
            return ''
